/**********************************************************************
 *<
	FILE: BMTEX.CPP

	DESCRIPTION: BMTEX 2D Texture map.

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include "stdmat.h"
#include <bmmlib.h>
#include "texutil.h"

extern HINSTANCE hInstance;

// Image filtering types
#define FILTER_PYR     0
#define FILTER_SAT     1
#define FILTER_NADA	   2

// Alpha source types
#define ALPHA_FILE 	0
#define ALPHA_RGB	2
#define ALPHA_NONE	3

// End conditions:
#define END_LOOP     0
#define END_PINGPONG 1
#define END_HOLD     2

#define NCOLS 0

class BMTex;

static Class_ID bmTexClassID(BMTEX_CLASS_ID,0);
static AColor black(0.0f,0.0f,0.0f,0.0f);
static inline AColor AColorFrom64(BMM_Color_64 pix) {
	return AColor((float)pix.r/65536.0f,(float)pix.g/65536.0f,(float)pix.b/65536.0f,(float)pix.a/65536.0f);
	}

static void LoadFailedMsg(const TCHAR* name) {
	if (name==NULL) return;
	TCHAR msg[128];
	wsprintf(msg,GetString(IDS_DS_FILE_NOT_FOUND), name);
	MessageBox(NULL, msg, GetString(IDS_DS_BITMAP_TEXTURE_ERR), MB_TASKMODAL);
	}

static void InvalidNameMsg(const TCHAR* name) {
	if (name==NULL) return;
	TCHAR msg[128];
	wsprintf(msg,_T("Invalid File Name: %s"), name);
	MessageBox(NULL, msg, GetString(IDS_DS_BITMAP_TEXTURE_ERR), MB_TASKMODAL);
	}

//------------------------------------------------------------------------
// BMSampler
//------------------------------------------------------------------------

class BMSampler: public MapSampler {
	Bitmap *bm;
	int alphaSource;
	public:
		BMSampler() { bm = NULL; }
		void Init(Bitmap *b, int asrc) { bm = b; alphaSource = asrc; }
		AColor Sample(ShadeContext& sc, float u,float v);
		AColor SampleFilter(ShadeContext& sc, float u,float v, float du, float dv);
//		float SampleMono(ShadeContext& sc, float u,float v);
//		float SampleMonoFilter(ShadeContext& sc, float u,float v, float du, float dv);
	} ;

AColor BMSampler::Sample(ShadeContext& sc, float u,float v) {
	BMM_Color_64 c;
	int w = bm->Width();
	int h = bm->Height();
	int x = (int)(frac(u)*(float)w);
	int y = h-1 - (int)((frac(v)*(float)h));
	bm->GetLinearPixels(x,y,1,&c);
	switch(alphaSource) {
		case ALPHA_NONE:  c.a = 0xffff; break;
		case ALPHA_RGB:   c.a = (c.r+c.g+c.b)/3; break;
		//  TBD
		// XPCOL needs to be handled in bitmap for filtering. 
		// Need to open a bitmap with this property.
		//	case ALPHA_XPCOL:  break; 
		}
	return AColorFrom64(c);
	}

AColor BMSampler::SampleFilter(ShadeContext& sc, float u,float v, float du, float dv) {
	BMM_Color_64 c;
	bm->GetFiltered(u,1.0f-v,du,dv,&c);
	switch(alphaSource) {
		case ALPHA_NONE:  c.a = 0xffff; break;
		case ALPHA_RGB:   c.a = (c.r+c.g+c.b)/3; break;
		//  TBD
		// XPCOL needs to be handled in bitmap for filtering. 
		// Need to open a bitmap with this property.
		//	case ALPHA_XPCOL:  break; 
		}
	return AColorFrom64(c);
	}


//------------------------------------------------------------------------
// BMAlphaSampler
//------------------------------------------------------------------------
class BMAlphaSampler: public MapSampler {
	Bitmap *bm;
	public:
		BMAlphaSampler() { bm = NULL; }
		void Init(Bitmap *b) { bm = b; }
		AColor Sample(ShadeContext& sc, float u,float v) { return AColor(0,0,0,0);}
		AColor SampleFilter(ShadeContext& sc, float u,float v, float du, float dv) { return AColor(0,0,0,0);}
		float SampleMono(ShadeContext& sc, float u,float v);
		float SampleMonoFilter(ShadeContext& sc, float u,float v, float du, float dv);
	} ;

float BMAlphaSampler::SampleMono(ShadeContext& sc, float u,float v) {
	BMM_Color_64 c;
	int w = bm->Width();
	int h = bm->Height();
	int x = (int)(frac(u)*(float)w);
	int y = h-1 - (int)(frac(v)*(float)h);
	bm->GetLinearPixels(x,y,1,&c);
	return( (float)c.a/65536.0f);
	}

float BMAlphaSampler::SampleMonoFilter(ShadeContext& sc, float u,float v, float du, float dv) {
	BMM_Color_64 c;
	bm->GetFiltered(u,1.0f-v,du,dv,&c);
	return( (float)c.a/65536.0f);
	}

//------------------------------------------------------------------------
// BMTexDlg
//------------------------------------------------------------------------
class BMTexDlg: public ParamDlg {
	public:
		HWND hwmedit;	 // window handle of the materials editor dialog
		IMtlParams *ip;
		BMTex *theBMTex;	 // current BMTex being edited.
		HWND hPanel; // Rollup pane
		HWND hTime; // Time Rollup pane
		TimeValue curTime; 
		ParamDlg *uvGenDlg;
		ParamDlg *texoutDlg;
		ISpinnerControl *startSpin,*rateSpin;
		BOOL valid;
		BOOL isActive;

		//-----------------------------
		BMTexDlg(HWND hwMtlEdit, IMtlParams *imp, BMTex *m); 
		~BMTexDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		BOOL TimeProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void HandleNameButton();
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff) {}
		void StuffBMNameField(HWND hwndDlg);
		void EnableAlphaButtons(BOOL isNew=FALSE);
		void Invalidate() { valid = FALSE;	InvalidateRect(hPanel,NULL,0); }

		// methods inherited from ParamDlg:
		Class_ID ClassID() {return bmTexClassID;  }
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theBMTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};


class BMTexPostLoad;

//--------------------------------------------------------------
// BMTex: A 2D texture map
//--------------------------------------------------------------

class BMTex: public BitmapTex {
	friend class BMTexPostLoad;
	friend class BMTexDlg;
	UVGen *uvGen;		   // ref #0
	IParamBlock *pblock;   // ref #1
	TextureOutput *texout; // ref #2
	BitmapInfo bi;
	TSTR bmName; // for loading old files only
	Bitmap *thebm;
	TexHandle *texHandle;
	float pbRate;
	TimeValue startTime;
	Interval ivalid;
	BMSampler mysamp;
	BMAlphaSampler alphasamp;
	BOOL loadingOld;
	int filterType;
	int alphaSource;
	int rollScroll;
	int endCond;
	int alphaAsMono;
	BOOL isNew;
	BOOL loadFailed; 
	BOOL inRender;
	BMTexDlg *paramDlg;
	int texTime;
	public:
		BMTex();
		~BMTex();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }
		TSTR GetFullName();

		void SetOutputLevel(TimeValue t, float v) {texout->SetOutputLevel(t,v); }
		void SetFilterType(int ft);
		void SetAlphaSource(int as);
		void SetEndCondition(int endcond) { endCond = endcond; }
		void SetAlphaAsMono(BOOL onoff) { alphaAsMono = onoff; }
		void SetMapName(TCHAR *name) { bi.SetName(name); }
		void SetStartTime(TimeValue t) { startTime = t; }
		void SetPlaybackRate(float r) { pbRate = r; }

		int GetFilterType() { return filterType; }
		int GetAlphaSource() { return alphaSource; }
		int GetEndCondition() { return endCond; }
		BOOL GetAlphaAsMono(BOOL onoff) { return alphaAsMono; }
		TCHAR *GetMapName() { return (TCHAR *)bi.Name(); }
		TimeValue GetStartTime() { return startTime; }
		float GetPlaybackRate() { return pbRate; }
		StdUVGen* GetUVGen() { return (StdUVGen*)uvGen; }
		TextureOutput* GetTexout() { return texout; }

		void NotifyChanged();
		void FreeBitmap();
 		BMMRES LoadBitmap(TimeValue t);
		int CalcFrame(TimeValue t);
		void ScaleBitmapBumpAmt(float f);
		void ReloadBitmap();

		// Evaluate the color of map for the context.
		RGBA EvalColor(ShadeContext& sc);
		float EvalMono(ShadeContext& sc);
		Point3 EvalNormalPerturb(ShadeContext& sc);

		void DiscardTexHandle();

		BOOL SupportTexDisplay() { return TRUE; }
		void ActivateTexDisplay(BOOL onoff);
		DWORD GetActiveTexHandle(TimeValue t, TexHandleMaker& thmaker);
		void GetUVTransform(Matrix3 &uvtrans) { uvGen->GetUVTransform(uvtrans); }
		int GetTextureTiling() { return  uvGen->GetTextureTiling(); }

		int RenderBegin(TimeValue t, ULONG flags) {	inRender = TRUE;return 1; 	}
		int RenderEnd(TimeValue t) { 	inRender = FALSE;	return 1; 	}

		int LoadMapFiles(TimeValue t) {	LoadBitmap(t);	return 1;	}

		Class_ID ClassID() {	return bmTexClassID; }
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_BITMAP); }  
		void DeleteThis() { delete this; }	

		// Requirements
		ULONG LocalRequirements(int subMtlNum) {
			return uvGen->Requirements(subMtlNum); 
			}

		int NumSubs() { return 2; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }
		void InitSlotType(int sType) { if (uvGen) uvGen->InitSlotType(sType); }

		// From ref
		int NumRefs() { return loadingOld?3:2; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		void EnumAuxFiles(NameEnumCallback& nameEnum, DWORD flags) {
			bi.EnumAuxFiles(nameEnum,flags);
			}
		int SetProperty(ULONG id, void *data);

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

	};

int numBMTexs = 0;
class BMTexClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new BMTex; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_BITMAP); }
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID() { return bmTexClassID; }
	const TCHAR* 	Category() { return TEXMAP_CAT_2D;  }
	};

static BMTexClassDesc bmTexCD;

ClassDesc* GetBMTexDesc() { return &bmTexCD;  }

static BOOL CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	BMTexDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (BMTexDlg*)lParam;
		theDlg->hPanel = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (BMTexDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = TRUE;
	int	res = theDlg->PanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = FALSE;
	return res;
	}

static BOOL CALLBACK  TimeDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	BMTexDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (BMTexDlg*)lParam;
		theDlg->hTime = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (BMTexDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = TRUE;
	int	res = theDlg->TimeProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = FALSE;
	return res;
	}

BMTexDlg::BMTexDlg(HWND hwMtlEdit, IMtlParams *imp, BMTex *m) { 
	hwmedit = hwMtlEdit;
	ip = imp;
	hPanel = NULL;
	theBMTex = m; 
	valid  = FALSE;
	isActive  = FALSE;
	uvGenDlg = theBMTex->uvGen->CreateParamDlg(hwmedit, imp);
	hPanel = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_BMTEX),
		PanelDlgProc, 
		GetString(IDS_DS_BITMAP_PARAMS), 
		(LPARAM)this );		
	texoutDlg = theBMTex->texout->CreateParamDlg(hwmedit, imp);
	hTime = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_BMTEX_TIME),
		TimeDlgProc, 
		GetString(IDS_DS_TIME_PARAMS), 
		(LPARAM)this,
		APPENDROLL_CLOSED );		
	curTime = imp->GetTime();
	}

void BMTexDlg::ReloadDialog() {
	Interval valid;
	theBMTex->Update(curTime, valid);
	LoadDialog(FALSE);
	}

void BMTexDlg::SetTime(TimeValue t) {
	Interval valid;
	uvGenDlg->SetTime(t);
	texoutDlg->SetTime(t);
	if (t!=curTime) {		
		curTime = t;
		theBMTex->Update(curTime, valid);
		LoadDialog(FALSE);
		InvalidateRect(hPanel,NULL,0);
		InvalidateRect(hTime,NULL,0);
		}
	}

BMTexDlg::~BMTexDlg() {
	theBMTex->paramDlg = NULL;
	SetWindowLong(hPanel, GWL_USERDATA, NULL);
	SetWindowLong(hTime, GWL_USERDATA, NULL);
	uvGenDlg->DeleteThis();
 	texoutDlg->DeleteThis();
	ReleaseISpinner(startSpin);
	ReleaseISpinner(rateSpin);
	hPanel =  NULL;
	hTime = NULL;
	}

void BMTex::FreeBitmap() {
	if (thebm) {
		delete thebm;
		thebm = NULL;
		}
	if (texHandle) {
		texHandle->DeleteThis();
		texHandle = NULL;	
		}
	loadFailed = FALSE;
	}

void BMTexDlg::HandleNameButton() {
	BOOL silent = TheManager->SetSilentMode(TRUE);
	BOOL res = TheManager->SelectFileInputEx(&theBMTex->bi, hPanel, GetString(IDS_DS_SELECT_BMFILE));
	TheManager->SetSilentMode(silent);
	if (res) {
		theBMTex->FreeBitmap();
		theBMTex->NotifyChanged();
		StuffBMNameField(hPanel);
		theBMTex->loadFailed = FALSE;
		BMMRES res = theBMTex->LoadBitmap(ip->GetTime());
		if (res==BMMRES_NODRIVER) 
		   	InvalidNameMsg(theBMTex->bi.Name());
		EnableAlphaButtons(TRUE);
	    UpdateMtlDisplay();
		}
	}

void GetBMName(BitmapInfo& bi, TSTR &fname) { 
	TSTR fullName;
	if (bi.Name()[0]==0)
		fullName = bi.Device();
	else 
		fullName =  bi.Name();
	SplitPathFile(fullName,NULL,&fname);
	}

void BMTexDlg::StuffBMNameField(HWND hwndDlg) {
	TSTR fname;
	GetBMName(theBMTex->bi,fname);
	SetDlgItemText(hPanel, IDC_BMTEX_NAME, fname);
	}

static int colID[2] = { IDC_CHECK_COL1, IDC_CHECK_COL2 };

void BMTexDlg::EnableAlphaButtons(BOOL isNew) {
	if (hPanel==NULL) return;
	if (theBMTex->bi.Flags()&MAP_HAS_ALPHA) {
		if (isNew) {
 			theBMTex->SetAlphaSource(ALPHA_FILE);
 			//theBMTex->SetAlphaAsMono(TRUE);
			}
		EnableWindow(GetDlgItem(hPanel,IDC_ALPHA_FILE), 1);
		}
	else {
		if (isNew||theBMTex->alphaSource==ALPHA_FILE) 
 			theBMTex->SetAlphaSource(ALPHA_NONE);
		EnableWindow(GetDlgItem(hPanel,IDC_ALPHA_FILE), 0);
		}
	CheckRadioButton( hPanel, IDC_ALPHA_FILE, IDC_ALPHA_NONE, theBMTex->alphaSource+IDC_ALPHA_FILE);
	}

BOOL BMTexDlg::PanelProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			StuffBMNameField(hwndDlg);

			EnableAlphaButtons();
			CheckRadioButton( hwndDlg, IDC_ALPHA_FILE, IDC_ALPHA_NONE, IDC_ALPHA_FILE+theBMTex->alphaSource);
			CheckRadioButton( hwndDlg, IDC_FILTER_PYR, IDC_FILTER_NADA, IDC_FILTER_PYR+theBMTex->filterType);
			CheckRadioButton( hwndDlg, IDC_BMTEX_RGBOUT, IDC_BMTEX_ALPHAOUT, IDC_BMTEX_RGBOUT+theBMTex->alphaAsMono);
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_BMTEX_NAME: 
					HandleNameButton();
					break;
				case IDC_ALPHA_FILE:
				case IDC_ALPHA_XPCOL:
				case IDC_ALPHA_RGB:
				case IDC_ALPHA_NONE:
					CheckRadioButton( hwndDlg, IDC_ALPHA_FILE, IDC_ALPHA_NONE, id);
					theBMTex->SetAlphaSource(id-IDC_ALPHA_FILE);
					theBMTex->NotifyChanged();
					break;
				case IDC_FILTER_NADA:
				case IDC_FILTER_PYR:
				case IDC_FILTER_SAT:
					CheckRadioButton( hwndDlg, IDC_FILTER_PYR, IDC_FILTER_NADA, id);
					theBMTex->SetFilterType(id-IDC_FILTER_PYR);
					theBMTex->NotifyChanged();
					break;
				case IDC_BMTEX_RELOAD:
					theBMTex->ReloadBitmap();
				    UpdateMtlDisplay(); // redraw viewports
					break;
				case IDC_BMTEX_RGBOUT:
				case IDC_BMTEX_ALPHAOUT:
					theBMTex->alphaAsMono = (id==IDC_BMTEX_ALPHAOUT)?1:0;
					theBMTex->NotifyChanged();
					break;
				}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case WM_PAINT: 	
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			break;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;		
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			theBMTex->NotifyChanged();
		    UpdateMtlDisplay();
			break;

    	}
	return FALSE;
	}


BOOL BMTexDlg::TimeProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			startSpin = SetupIntSpinner(hwndDlg, IDC_BMTEX_START_SPIN, IDC_BMTEX_START,-100000 ,100000, 
				theBMTex->startTime/GetTicksPerFrame());
			rateSpin = SetupFloatSpinner(hwndDlg, IDC_BMTEX_RATE_SPIN, IDC_BMTEX_RATE,0.0f, 1000.0f, 1.0f,.1f);
			CheckRadioButton( hwndDlg, IDC_BMTEX_LOOP, IDC_BMTEX_HOLD, IDC_BMTEX_LOOP+theBMTex->endCond);
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_BMTEX_LOOP:
				case IDC_BMTEX_HOLD:
				case IDC_BMTEX_PINGPONG:
					theBMTex->endCond = id-IDC_BMTEX_LOOP;
					break;
				}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (id) {
				case IDC_BMTEX_START_SPIN: 
					theBMTex->startTime = startSpin->GetIVal()*GetTicksPerFrame(); 	
					break;
				case IDC_BMTEX_RATE_SPIN: 
					theBMTex->pbRate = rateSpin->GetFVal(); 	
					break;
				}
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;		
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) 
				theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else 
				theHold.Cancel();
			theBMTex->NotifyChanged();
		    UpdateMtlDisplay();
			break;

    	}
	return FALSE;
	}


void BMTexDlg::LoadDialog(BOOL draw) {
	if (theBMTex) {
		Interval valid;
		theBMTex->Update(curTime,valid);
		StuffBMNameField(hPanel);
		CheckRadioButton( hPanel, IDC_ALPHA_FILE, IDC_ALPHA_NONE, IDC_ALPHA_FILE+theBMTex->alphaSource);
		CheckRadioButton( hPanel, IDC_FILTER_PYR, IDC_FILTER_NADA, IDC_FILTER_PYR+theBMTex->filterType);
		CheckRadioButton( hPanel, IDC_BMTEX_RGBOUT, IDC_BMTEX_ALPHAOUT, IDC_BMTEX_RGBOUT+theBMTex->alphaAsMono);

		startSpin->SetValue(theBMTex->startTime/GetTicksPerFrame(),FALSE);
		rateSpin->SetValue(theBMTex->pbRate,FALSE);
		CheckRadioButton( hTime, IDC_BMTEX_LOOP, IDC_BMTEX_HOLD, IDC_BMTEX_LOOP+theBMTex->endCond);
		}
	}

void BMTexDlg::SetThing(ReferenceTarget *m) {
	assert (m->ClassID()==bmTexClassID);
	assert (m->SuperClassID()==TEXMAP_CLASS_ID);
	if (theBMTex) theBMTex->paramDlg = NULL;
	theBMTex = (BMTex *)m;
	if (theBMTex) theBMTex->paramDlg = this;
	uvGenDlg->SetThing(theBMTex->uvGen);
	texoutDlg->SetThing(theBMTex->texout);
	LoadDialog(TRUE);
	}


//-----------------------------------------------------------------------------
//  BMTex
//-----------------------------------------------------------------------------


#define BMTEX_VERSION 1

#define PB_BLUR	0
#define NPARAMS 1

static ParamBlockDescID pbdesc1[] = {
	{ TYPE_FLOAT, NULL, TRUE,1} 	// blur
	};   

static int name_id[NPARAMS] = { IDS_DS_BLUR };

void BMTex::Reset() {
	if (uvGen) uvGen->Reset();
	else ReplaceReference( 0, GetNewDefaultUVGen());	
	if (texout) texout->Reset();
	else ReplaceReference( 1, GetNewDefaultTextureOutput());	
	filterType = FILTER_PYR;
	alphaSource = ALPHA_FILE;
	alphaAsMono = FALSE;
	ivalid.SetEmpty();
	if (!isNew) {
		FreeBitmap();
		NotifyChanged();
	    if (paramDlg) paramDlg->UpdateMtlDisplay();
		}
	isNew = FALSE;
	}

void BMTex::NotifyChanged() {
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	}

BMTex::BMTex() {
	paramDlg = NULL;
	pblock = NULL;
	uvGen = NULL;
	texout = NULL;
	thebm = NULL;
	texHandle = NULL;
	loadFailed = FALSE;
	inRender = FALSE;
	startTime = 0;
	pbRate = 1.0f;
	endCond = END_LOOP;
	isNew = TRUE;
	loadingOld = FALSE;
	Reset();
	rollScroll=0;
	}

static int bmFilterType(int filt_type) {
	switch(filt_type) {
		case FILTER_PYR: return BMM_FILTER_PYRAMID;
		case FILTER_SAT: return BMM_FILTER_SUM;
		default: return BMM_FILTER_NONE;
		}
	}

void BMTex::SetFilterType(int ft) {
	filterType = ft;
	if (thebm)	
		thebm->SetFilter(bmFilterType(filterType));
	}

void BMTex::SetAlphaSource(int as) {
	if (as!=alphaSource) {
		alphaSource =as; 
		mysamp.Init(thebm, alphaSource);
		alphasamp.Init(thebm);
		}
	}

static TimeValue modt(TimeValue x, TimeValue m) {
	TimeValue n = (int)(x/m);
	x -= n*m;
	return (x<0)? x+m : x ;
	}

int BMTex::CalcFrame(TimeValue t) {
	TimeValue tm,dur,td;
	int fstart = bi.FirstFrame();
	int fend = bi.LastFrame();
	int tpf = GetTicksPerFrame();
	tm = TimeValue(float(t-startTime)*pbRate);
	dur = (fend-fstart+1)*GetTicksPerFrame();
	switch (endCond) {
		case END_HOLD:
			if (tm<=0) return fstart;
			if (tm>=dur) return fend;
			return tm/tpf;
		case END_PINGPONG:
			if ((tm/dur)&1) {
				td = modt(tm,dur);
				return fstart + fend-td/tpf;
				}
			// else fall thrue--
		case END_LOOP:
			td = modt(tm,dur);
			return td/tpf;
		}
	return 0;
	}

BMMRES BMTex::LoadBitmap(TimeValue t) {
	BMMRES status = BMMRES_SUCCESS;
	if (bi.Name()[0]==0)
		if (bi.Device()[0]==0)
			return BMMRES_NODRIVER;
	if (thebm==NULL) {
		bi.SetCurrentFrame(CalcFrame(t));
		BOOL silent = TheManager->SetSilentMode(TRUE);
		SetCursor(LoadCursor(NULL,IDC_WAIT));
		thebm = TheManager->Load(&bi,&status);
		SetCursor(LoadCursor(NULL,IDC_ARROW));
		TheManager->SetSilentMode(silent);
		if (thebm==NULL) {
			return status;
			}
		thebm->SetFilter(bmFilterType(filterType));
		// fixup for imported files
		if (!(bi.Flags()&MAP_HAS_ALPHA)&&alphaSource==ALPHA_FILE) {
			if (paramDlg) 
				paramDlg->EnableAlphaButtons(FALSE);
			else 
	 			SetAlphaSource(ALPHA_NONE);
			}			
		}
//	else {
		bi.SetCurrentFrame(CalcFrame(t));
		BOOL silent = TheManager->SetSilentMode(TRUE);
		status = thebm->GoTo(&bi);
		TheManager->SetSilentMode(silent);
//		}
	mysamp.Init(thebm, alphaSource);
	alphasamp.Init(thebm);
	return status;
	}

void BMTex::ReloadBitmap() {
	if (thebm) {
		loadFailed = FALSE;
		FreeBitmap(); 
		LoadBitmap(paramDlg->ip->GetTime());
//		TheManager->LoadInto(&bi,&thebm);
		}
	else {
		loadFailed = FALSE;
		LoadBitmap(paramDlg->ip->GetTime());
		}
	paramDlg->EnableAlphaButtons(FALSE);
	NotifyChanged();
	}

void BMTex::DiscardTexHandle() {
	if (texHandle) {
		texHandle->DeleteThis();
		texHandle = NULL;
		}
	}

void BMTex::ActivateTexDisplay(BOOL onoff) {
	if (!onoff) 
		DiscardTexHandle();
	}

DWORD BMTex::GetActiveTexHandle(TimeValue t, TexHandleMaker& thmaker) {
	if (texHandle&&texTime==CalcFrame(t)) 
		return texHandle->GetHandle();
	else {
		DiscardTexHandle();  // DS 9/9/96
		LoadBitmap(t);
		if (thebm==NULL) {
			if (!loadFailed) {
				TSTR fname;
				GetBMName(bi,fname);
				LoadFailedMsg(fname.data());
				loadFailed = TRUE;
				}
			return 0;
			}
		texTime = CalcFrame(t);
		texHandle = thmaker.CreateHandle(thebm,uvGen->SymFlags());
		return texHandle->GetHandle();
		}
	}
							 
//----------------------------------------------------------- 

AColor BMTex::EvalColor(ShadeContext& sc) {
	if (gbufID) 
		sc.SetGBufferID(gbufID);
	if (!sc.doMaps) return AColor(0.0f,0.0f,0.0f,1.0f);
	if (thebm==NULL) 
		return black;
	return texout->Filter(uvGen->EvalUVMap(sc,&mysamp,filterType!=FILTER_NADA));
	}

float BMTex::EvalMono(ShadeContext& sc) {
	if (gbufID) sc.SetGBufferID(gbufID);
	if (thebm==NULL) 
		return black;
	if (alphaAsMono) 
		return texout->Filter(uvGen->EvalUVMapMono(sc,&alphasamp,filterType!=FILTER_NADA));
	else 
		return texout->Filter(uvGen->EvalUVMapMono(sc,&mysamp,filterType!=FILTER_NADA));
	}

Point3 BMTex::EvalNormalPerturb(ShadeContext& sc) {
	Point3 dPdu, dPdv;
	Point2 dM;
	if (gbufID) sc.SetGBufferID(gbufID);
	if (thebm==NULL) 
		return Point3(0,0,0);
	uvGen->GetBumpDP(sc,dPdu,dPdv);  // get bump basis vectors
	if (alphaAsMono) 
		dM =(.01f)*uvGen->EvalDeriv(sc,&alphasamp,filterType!=FILTER_NADA);
	else 
		dM =(.01f)*uvGen->EvalDeriv(sc,&mysamp,filterType!=FILTER_NADA);


#if 0
	// Blinn's algorithm
	Point3 N = sc.Normal();
	Point3 uVec = CrossProd(N,dPdv);
	Point3 vVec = CrossProd(N,dPdu);
	return texout->Filter(dM.x*uVec-dM.y*vVec);
#else						 
	return texout->Filter(dM.x*dPdu+dM.y*dPdv);
#endif
	}

RefTargetHandle BMTex::Clone(RemapDir &remap) {
	BMTex *mnew = new BMTex();
	*((MtlBase*)mnew) = *((MtlBase*)this);  // copy superclass stuff
	mnew->ReplaceReference(0,remap.CloneRef(uvGen));
	mnew->ReplaceReference(1,remap.CloneRef(texout));
	mnew->filterType = filterType;
	mnew->alphaSource = alphaSource;
	mnew->alphaAsMono  = alphaAsMono;
	mnew->endCond =  endCond;
	mnew->ivalid.SetEmpty();
	mnew->bi = bi;
	return (RefTargetHandle)mnew;
	}

ParamDlg* BMTex::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp) {
	BMTexDlg *dm = new BMTexDlg(hwMtlEdit, imp, this);
	dm->LoadDialog(TRUE);	
	paramDlg = dm;
	return dm;	
	}

BMTex::~BMTex() {
	DiscardTexHandle();
	FreeBitmap();
	}


void BMTex::Update(TimeValue t, Interval& valid) {		
	if (!ivalid.InInterval(t)) {
		ivalid.SetInfinite();
		uvGen->Update(t,ivalid);
		texout->Update(t,ivalid);
		}
	if (thebm) {
		if (bi.FirstFrame()!=bi.LastFrame())
			ivalid.SetInstant(t);  // force bitmap to be reloaded
		}
	else 
		ivalid.SetInstant(t);  // force bitmap to be reloaded 
	valid &= ivalid;
	}

RefTargetHandle BMTex::GetReference(int i) {
	switch(i) {
		case 0: return uvGen;
		case 1:	return texout ;
		default: return NULL;
		}
	}

void BMTex::SetReference(int i, RefTargetHandle rtarg) {
	if (loadingOld) {
		switch(i) {
			case 0: uvGen = (UVGen *)rtarg; break;
			case 2:	texout = (TextureOutput *)rtarg; break;
			default: break;
			}
		}
	else {
		switch(i) {
			case 0: uvGen = (UVGen *)rtarg; break;
			case 1:	texout = (TextureOutput *)rtarg; break;
			default: break;
			}
		}
	}

	 
Animatable* BMTex::SubAnim(int i) {
	switch (i) {
		case 0: return uvGen;
		case 1: return texout;
		default: assert(0); return NULL;
		}
	}


TSTR BMTex::GetFullName() {
	TSTR cnm,nm,fname;
   	GetClassName(cnm);
	GetBMName(bi,fname);
	nm.printf(_T("%s  ( %s ) %s"),GetName().data(), cnm.data(),fname.data());
	return nm;
	}

TSTR BMTex::SubAnimName(int i) {
	switch (i) {
		case 0: return TSTR(GetString(IDS_DS_COORDINATES));		
		case 1: return TSTR(GetString(IDS_DS_PARAMETERS));		
		case 2: return TSTR(GetString(IDS_DS_OUTPUT));		
		default: assert(0); return TSTR(_T(""));
		}
	}

RefResult BMTex::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
   PartID& partID, RefMessage message ) {
	switch (message) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			if (hTarget!=uvGen&&hTarget!=texout) {
				if (paramDlg&&!paramDlg->isActive) 
				 	paramDlg->Invalidate();
				}
			break;

		case REFMSG_UV_SYM_CHANGE:
			DiscardTexHandle();  
			break;

		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_BLUR: gpd->dim = stdPercentDim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			gpn->name= TSTR(GetString(name_id[gpn->index]));
			return REF_STOP; 
			}
		}
	return(REF_SUCCEED);
	}



//------------------------------------------------------------------------
// IO
//------------------------------------------------------------------------


class BMTexPostLoad : public PostLoadCallback {
	public:
		BMTex *chk;
		BMTexPostLoad(BMTex *b) {chk=b;}
		void proc(ILoad *iload) {
			chk->loadingOld = FALSE;
			if (chk->bmName.Length()>0) {
				chk->bi.SetName(chk->bmName);   // for obsolete files	
				iload->SetObsolete();
				}
			delete this;
			}
	};


#define MTL_HDR_CHUNK 0x4000
#define OLDBMTEX_NAME_CHUNK 0x5001
#define BMTEX_FILTER_CHUNK 0x5002
#define BMTEX_ALPHASOURCE_CHUNK 0x5003
#define BMTEX_NAME_CHUNK 0x5004
#define BMTEX_IO_CHUNK 0x5010
#define BMTEX_START_CHUNK 0x5011
#define BMTEX_RATE_CHUNK 0x5012
#define BMTEX_ALPHA_MONO_CHUNK 0x5013
#define BMTEX_ENDCOND_CHUNK 0x5014
#define BMTEX_VERS1_CHUNK 0x5021

IOResult BMTex::Save(ISave *isave) { 
	ULONG nb;
	IOResult res;
	// Save common stuff
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();

	isave->BeginChunk(BMTEX_FILTER_CHUNK);
	isave->Write(&filterType,sizeof(filterType),&nb);			
	isave->EndChunk();

	isave->BeginChunk(BMTEX_ALPHASOURCE_CHUNK);
	isave->Write(&alphaSource,sizeof(alphaSource),&nb);			
	isave->EndChunk();

	if (startTime!=0) {
		isave->BeginChunk(BMTEX_START_CHUNK);
		isave->Write(&startTime,sizeof(startTime),&nb);			
		isave->EndChunk();
		}
	if (pbRate!=1.0) {
		isave->BeginChunk(BMTEX_RATE_CHUNK);
		isave->Write(&pbRate,sizeof(pbRate),&nb);			
		isave->EndChunk();
		}
	if (alphaAsMono) {
		isave->BeginChunk(BMTEX_ALPHA_MONO_CHUNK);
		isave->EndChunk();
		}
	if (endCond != END_LOOP) {
		isave->BeginChunk(BMTEX_ENDCOND_CHUNK);
		isave->Write(&endCond,sizeof(endCond),&nb);			
		isave->EndChunk();
		}
	isave->BeginChunk(BMTEX_IO_CHUNK);
	bi.Save(isave);
	isave->EndChunk();

	isave->BeginChunk(BMTEX_VERS1_CHUNK);
	isave->EndChunk();


	return IO_OK;
	}	

IOResult BMTex::Load(ILoad *iload) { 
	ULONG nb;
	IOResult res;
	bmName.Resize(0);
	loadingOld = TRUE;
	iload->RegisterPostLoadCallback(new BMTexPostLoad(this));
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			case BMTEX_VERS1_CHUNK:
				loadingOld = FALSE;
				break;
			case BMTEX_FILTER_CHUNK:
				iload->Read(&filterType,sizeof(filterType),&nb);			
				break;
			case BMTEX_ALPHASOURCE_CHUNK:
				iload->Read(&alphaSource,sizeof(alphaSource),&nb);			
				break;
			case BMTEX_NAME_CHUNK:
				TCHAR *buf;
				if (IO_OK==iload->ReadWStringChunk(&buf)) 
					bmName = buf;					
				break;
			case BMTEX_IO_CHUNK:
				res = bi.Load(iload);
				// DS 7/18/96 - This is to handle old files 
				// that had the bi saved with start = end = 0, 
				// which now screws up animation.
				bi.SetStartFrame(BMM_UNDEF_FRAME);
				bi.SetEndFrame(BMM_UNDEF_FRAME);
				break;
			case BMTEX_START_CHUNK:
				iload->Read(&startTime,sizeof(startTime),&nb);			
				break;
			case BMTEX_RATE_CHUNK:
				iload->Read(&pbRate,sizeof(pbRate),&nb);			
				break;
			case BMTEX_ALPHA_MONO_CHUNK:
				alphaAsMono = TRUE;
				break;
			case BMTEX_ENDCOND_CHUNK:
				iload->Read(&endCond,sizeof(endCond),&nb);			
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	if (alphaSource<0||alphaSource>ALPHA_NONE) 
		alphaSource = ALPHA_FILE;
	return IO_OK;
	}

int BMTex::SetProperty(ULONG id, void *data)
	{
	switch (id) {
		case PROPID_CLEARCACHES:
			FreeBitmap();
			return 1;

		default: return 0;
		}
	}


