/**********************************************************************
 *<
	FILE: cmtl.cpp

	DESCRIPTION:  Composite material

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include "stdmtl.h"

extern HINSTANCE hInstance;

#define NSUBMTLS 2

static Class_ID cmtlClassID(CMTL_CLASS_ID,0);

class CMtl;

class CMtlDlg: public ParamDlg {
	public:
		HWND hwmedit;	 // window handle of the materials editor dialog
		IMtlParams *ip;
		CMtl *theMtl;	 // current mtl being edited.
		HWND hPanelBasic; // Rollup pane
		ISpinnerControl *ySpin,*wSpin;
		TimeValue curTime; 
		int isActive;
		BOOL valid;
		CMtlDlg(HWND hwMtlEdit, IMtlParams *imp, CMtl *m); 
		~CMtlDlg();
		BOOL BasicPanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void UpdateSubMtlNames();
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff) {}
		void Invalidate() { valid = FALSE;	InvalidateRect(hPanelBasic,NULL,0); }

		// methods inherited from ParamDlg:
		Class_ID ClassID() {return cmtlClassID;  }
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theMtl; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};

// Parameter block indices
#define PB_THRESH		0
#define PB_WIDTH		1

class CMtl: public Mtl {
	friend class CMtlDlg;
	// Animatable parameters
	IParamBlock *pblock;
	Interval ivalid;
	CMtlDlg *paramDlg;
	Mtl* subMtl[NSUBMTLS];
	float thresh;
	float width;
	public:

		void SetAmbient(Color c, TimeValue t) {}		
		void SetDiffuse(Color c, TimeValue t) {}		
		void SetSpecular(Color c, TimeValue t) {}
		void SetShininess(float v, TimeValue t) {}		
		void SetThresh(float v, TimeValue t);
		void SetWidth(float v, TimeValue t);
		float GetThresh() { return thresh; }
		float GetWidth() { return width; }

		Mtl *UseMtl();
		Color GetAmbient(int mtlNum=0, BOOL backFace=FALSE);
	    Color GetDiffuse(int mtlNum=0, BOOL backFace=FALSE);
		Color GetSpecular(int mtlNum=0, BOOL backFace=FALSE);
		float GetXParency(int mtlNum=0, BOOL backFace=FALSE);
		float GetShininess(int mtlNum=0, BOOL backFace=FALSE);		
		float GetShinStr(int mtlNum=0, BOOL backFace=FALSE);
		float WireSize(int mtlNum=0, BOOL backFace=FALSE);

		CMtl();
		void SetParamDlg( CMtlDlg *pd) { paramDlg = pd; }
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Shade(ShadeContext& sc);
		void Update(TimeValue t, Interval& valid);
		void SwapInputs();
		void Reset();
		Interval Validity(TimeValue t);
		void NotifyChanged();

		Class_ID ClassID() {	return cmtlClassID; }
		SClass_ID SuperClassID() { return MATERIAL_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_TOPBOTTOM); }  
		void CMtl::Fixup();  //TEMPORARY
			

		void DeleteThis() { delete this; }	

		// Methods to access sub-materials of meta-materials
	   	int NumSubMtls() { return NSUBMTLS; }
		Mtl* GetSubMtl(int i) { return subMtl[i]; }
		void SetSubMtl(int i, Mtl *m);
		TSTR GetSubMtlSlotName(int i);

		int NumSubs() { return NSUBMTLS+1; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return NSUBMTLS+1;  }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

	};


int numCMtls = 0;
class CMtlClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new CMtl; }
	const TCHAR *	ClassName() { return  GetString(IDS_DS_TOPBOTTOM); }  
	SClass_ID		SuperClassID() { return MATERIAL_CLASS_ID; }
	Class_ID 		ClassID() { return cmtlClassID; }
	const TCHAR* 	Category() { return _T("");  }
	};

static CMtlClassDesc cMtlCD;

ClassDesc* GetCMtlDesc() { 
	return &cMtlCD;  
	}

static BOOL CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	CMtlDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (CMtlDlg*)lParam;
		theDlg->hPanelBasic = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (CMtlDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = 1;
	int	res = theDlg->BasicPanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = 0;
	return res;
	}



CMtlDlg::CMtlDlg(HWND hwMtlEdit, IMtlParams *imp, CMtl *m) { 
	hwmedit = hwMtlEdit;
	ip = imp;
	hPanelBasic = NULL;
	theMtl = m; 
	isActive = 0;
	valid = FALSE;
	hPanelBasic = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_CMTL),
		PanelDlgProc, 
		GetString(IDS_DS_BASIC), 
		(LPARAM)this );		
	curTime = imp->GetTime();
	}

void CMtlDlg::ReloadDialog() {
	Interval valid;
	theMtl->Update(curTime,valid);
	LoadDialog(FALSE);
	}

void CMtlDlg::SetTime(TimeValue t) {
	if (t!=curTime) {
		Interval valid;
		curTime = t;
		theMtl->Update(curTime,valid);
		LoadDialog(FALSE);
		InvalidateRect(hPanelBasic,NULL,0);
		}
	}

CMtlDlg::~CMtlDlg() {
	theMtl->SetParamDlg(NULL);
	ReleaseISpinner(ySpin);
	ReleaseISpinner(wSpin);
	SetWindowLong(hPanelBasic, GWL_USERDATA, NULL);
	}


static int ThreshToInt(float f) { return (int)((f+1.0f)*50.0f); }
static float IntToThresh(int i) { return (float(i)/50.0f-1.0f); }

static int WidthToInt(float f) { return (int)((f)*100.0f); }
static float IntToWidth(int i) { return (float(i)/100.0f); }


BOOL CMtlDlg::BasicPanelProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			ySpin = SetupIntSpinner(hwndDlg, IDC_CMTL_Y_SPIN, IDC_CMTL_Y_EDIT,
				0,100,ThreshToInt(theMtl->GetThresh()));
			wSpin = SetupIntSpinner(hwndDlg, IDC_CMTL_W_SPIN, IDC_CMTL_W_EDIT,
			 0,100,WidthToInt(theMtl->GetWidth()));
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_TOP_MTL: 
					PostMessage(hwmedit,WM_SUB_MTL_BUTTON,0 ,(LPARAM)theMtl);
					break;
				case IDC_BOT_MTL: 
					PostMessage(hwmedit,WM_SUB_MTL_BUTTON,1 ,(LPARAM)theMtl);
					break;
				case IDC_CMTL_SWAP:
					theMtl->SwapInputs();				
					LoadDialog(TRUE);
					break; 
				}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case WM_PAINT: 
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			break;
		case WM_CLOSE:
			break;       
		case WM_DESTROY: 
			break;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (id) {
				case IDC_CMTL_Y_SPIN: 
					theMtl->SetThresh(IntToThresh(ySpin->GetIVal()),curTime); 
					break;
				case IDC_CMTL_W_SPIN: 
					theMtl->SetWidth(IntToWidth(wSpin->GetIVal()),curTime); 
					break;
				}
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
		    UpdateMtlDisplay();
			break;

    	}
	return FALSE;
	}

static BOOL CALLBACK  BasicPanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	CMtlDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (CMtlDlg*)lParam;
		theDlg->hPanelBasic = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (CMtlDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = 1;
	int	res = theDlg->BasicPanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = 0;
	return res;
	}

static int subMtlId[NSUBMTLS] = { IDC_TOP_MTL, IDC_BOT_MTL };

void CMtlDlg::UpdateSubMtlNames() {
	for (int i=0; i<NSUBMTLS; i++) {
		Mtl *m = theMtl->subMtl[i];
		TSTR nm;
		if (m) 
			nm = m->GetFullName();
		else 
			nm = GetString(IDS_DS_NONE);
		SetDlgItemText(hPanelBasic, subMtlId[i], nm.data());
		}
	}

void CMtlDlg::LoadDialog(BOOL draw) {
	if (theMtl) {
		Interval valid;
		theMtl->Update(curTime,valid);
		ySpin->SetValue(ThreshToInt(theMtl->GetThresh()),FALSE);
		wSpin->SetValue(WidthToInt(theMtl->GetWidth()),FALSE);
		UpdateSubMtlNames();
		}
	}

void CMtlDlg::SetThing(ReferenceTarget *m) {
	assert (m->ClassID()==cmtlClassID);
	assert (m->SuperClassID()==MATERIAL_CLASS_ID);
	if (theMtl) theMtl->paramDlg = NULL;
	theMtl = (CMtl *)m;
	if (theMtl) theMtl->paramDlg = this;
	LoadDialog(TRUE);
	}

//-----------------------------------------------------------------------------
//  CMtl
//-----------------------------------------------------------------------------

static ParamBlockDesc pbdesc[] = {
	{ TYPE_FLOAT, NULL, TRUE },   // thresh
	{ TYPE_FLOAT, NULL, TRUE } };   // blend

void CMtl::Reset() {
	ReplaceReference( 0, CreateParameterBlock( pbdesc, 2 ) );	
	ivalid.SetEmpty();
	//DeleteReference(1);
	//DeleteReference(2);
	ReplaceReference(1,new StdMtl);
	ReplaceReference(2,new StdMtl);
	thresh = 0.0f;
	width = .05f;
	}

void CMtl::NotifyChanged() {
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	}

CMtl::CMtl() {
	subMtl[0] = subMtl[1] = NULL;
	pblock = NULL;
	paramDlg = NULL;
	Reset();
	}

void CMtl::SwapInputs() {
	Mtl *tmp = subMtl[0];
	subMtl[0] = subMtl[1];
	subMtl[1] = tmp;
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	NotifyDependents(FOREVER,PART_ALL,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	}

Mtl *CMtl::UseMtl() {
	Mtl* m = NULL;
	for (int i=0; i<NSUBMTLS; i++) if (subMtl[i]) { m = subMtl[i]; break; }
	return m;
	}

// These allow the real-time renderer to display a material appearance.
// Just use the "Top material"
Color CMtl::GetAmbient(int mtlNum, BOOL backFace) { 
	return UseMtl()?UseMtl()->GetAmbient(mtlNum,backFace):Color(0,0,0);
	}		
Color CMtl::GetDiffuse(int mtlNum, BOOL backFace){ 
	return UseMtl()?UseMtl()->GetDiffuse(mtlNum,backFace):Color(0,0,0);
	}				
Color CMtl::GetSpecular(int mtlNum, BOOL backFace){
	return UseMtl()?UseMtl()->GetSpecular(mtlNum,backFace):Color(0,0,0);
	}		
float CMtl::GetXParency(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}
float CMtl::GetShininess(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}		
float CMtl::GetShinStr(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}
float CMtl::WireSize(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->WireSize(mtlNum,backFace):0.0f;
	}


RefTargetHandle CMtl::Clone(RemapDir &remap) {
	CMtl *mnew = new CMtl();
	*((MtlBase*)mnew) = *((MtlBase*)this);  // copy superclass stuff
	mnew->ReplaceReference(0,remap.CloneRef(pblock));
	mnew->thresh = thresh;
	mnew->width = width;
	mnew->ivalid.SetEmpty();	
	for (int i = 0; i<NSUBMTLS; i++) {
		if (subMtl[i])
			mnew->ReplaceReference(i+1,remap.CloneRef(subMtl[i]));
		}
	return (RefTargetHandle)mnew;
	}

	
ParamDlg* CMtl::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp) {
	CMtlDlg *dm = new CMtlDlg(hwMtlEdit, imp, this);
	dm->LoadDialog(TRUE);	
	SetParamDlg(dm);
	return dm;	
	}

void CMtl::Update(TimeValue t, Interval& valid) {		
	if (!ivalid.InInterval(t)) {
		ivalid.SetInfinite();
		pblock->GetValue( PB_THRESH, t, thresh, ivalid );
		pblock->GetValue( PB_WIDTH, t, width, ivalid );
		for (int i=0; i<NSUBMTLS; i++) {
			if (subMtl[i]) 
				subMtl[i]->Update(t,ivalid);
			}
		}
	valid &=ivalid;
	}

Interval CMtl::Validity(TimeValue t) {
	Interval valid;
	Update(t,valid);
	return ivalid;
	}

void CMtl::SetThresh(float v, TimeValue t) {
	thresh = v;
	pblock->SetValue( PB_THRESH, t, v);
	NotifyChanged();
	}

void CMtl::SetWidth(float v, TimeValue t) {
	width = v;
	pblock->SetValue( PB_WIDTH, t, v);
	NotifyChanged();
	}

RefTargetHandle CMtl::GetReference(int i) {
	if (i==0)
		return pblock ;
	else 
		return subMtl[i-1];
	}

void CMtl::SetReference(int i, RefTargetHandle rtarg) {
	if (i==0)
		pblock = (IParamBlock *)rtarg;
	else 
		subMtl[i-1] = (Mtl *)rtarg;
	}

void CMtl::SetSubMtl(int i, Mtl *m) {
	ReplaceReference(i+1,m);				  
	if (paramDlg)
		paramDlg->UpdateSubMtlNames();
	}

TSTR CMtl::GetSubMtlSlotName(int i) {
	switch(i) {
		case 0:  return TSTR(GetString(IDS_DS_TOP)); 
		case 1:  return TSTR(GetString(IDS_DS_BOTTOM)); 
		default: return TSTR(_T(""));
		}
	}
	 
Animatable* CMtl::SubAnim(int i) {
	if (i==0) return pblock;
	else  return subMtl[i-1]; 
	}

TSTR CMtl::SubAnimName(int i) {
	if (i==0) return TSTR(GetString(IDS_DS_PARAMETERS));		
	else 
		return GetSubMtlTVName(i-1);
	}

RefResult CMtl::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
   PartID& partID, RefMessage message ) {
	switch (message) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			if (paramDlg&&!paramDlg->isActive) {
				paramDlg->Invalidate();
				}
			break;

		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_THRESH: gpd->dim = stdPercentDim; break;
				case PB_WIDTH: gpd->dim = stdPercentDim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			switch (gpn->index) {
				case PB_THRESH: gpn->name = TSTR(GetString(IDS_DS_THRESH)); break;
				case PB_WIDTH: gpn->name = TSTR(GetString(IDS_DS_BLEND)); break;
				}
			return REF_STOP; 
			}
		}
	return(REF_SUCCEED);
	}


inline void Clamp(Color &c) {
	if (c.r > 1.0f) c.r = 1.0f;
	if (c.g > 1.0f) c.g = 1.0f;
	if (c.b > 1.0f) c.b = 1.0f;
	}
static Color black(0,0,0);

void CMtl::Shade(ShadeContext& sc) {
	Point3 N = sc.Normal();
	if (gbufID) sc.SetGBufferID(gbufID);
	N = sc.VectorTo(N, REF_WORLD);
	if (N.z>thresh+width) {
		if (subMtl[0]) subMtl[0]->Shade(sc);
		}
	else if (N.z<thresh-width) {
		if (subMtl[1]) subMtl[1]->Shade(sc);
		}
	else {
		if (subMtl[0]==NULL) {
			if (subMtl[1]) subMtl[1]->Shade(sc);
			}
		else 
		if (subMtl[1]==NULL) {
			if (subMtl[0]) subMtl[0]->Shade(sc);
			}
		else {
			float b = (N.z-thresh+width)/(2.0f*width);
			subMtl[0]->Shade(sc);
			ShadeOutput out0 = sc.out;
			sc.ResetOutput();
			subMtl[1]->Shade(sc);
			sc.out.MixIn(out0,1.0f-b);
			}
		}
	}

void CMtl::Fixup() {	
	// Old style: have to fix up the references since they changed.
	SetReference(2,GetReference(1));
	SetReference(1,GetReference(0));
	SetReference(0,NULL);
	ReplaceReference( 0, CreateParameterBlock( pbdesc, 2 ) );	
	SetThresh(thresh,0);
	SetWidth(width,0);
	}

#define CMTL_YTHRESH_CHUNK 0x1111
#define CMTL_YWIDTH_CHUNK 0x1112
#define MTL_HDR_CHUNK 0x4000

IOResult CMtl::Save(ISave *isave) { 
	IOResult res;
	
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();
	return IO_OK;
	}	
	  
class CMtlCB: public PostLoadCallback {
	CMtl *mtl;
	public:
		CMtlCB(CMtl *m) { mtl = m; }
		void proc(ILoad *iload) {
			mtl->Fixup();
			delete this;
			}
	};



IOResult CMtl::Load(ILoad *iload) { 
	ULONG nb;
	IOResult res;
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			case CMTL_YTHRESH_CHUNK:
				res = iload->Read(&thresh,sizeof(float), &nb);
				iload->SetObsolete();
				iload->RegisterPostLoadCallback(new CMtlCB(this));
				break;
			case CMTL_YWIDTH_CHUNK:
				res = iload->Read(&width,sizeof(float), &nb);
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	return IO_OK;
                                                      
	}
