/**********************************************************************
 *<
	FILE:			Dent.h

	DESCRIPTION:	Dent 3D Texture map class decl.

	CREATED BY:		Suryan Stalin, on 4th April 1996

	HISTORY:		Modified from Marble.cpp by adding IPAS dent stuff
	*>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/


#ifndef __DENT__H
#define __DENT__H


//	defines
#define DENT_CLASS_ID 		0x0000218
#define NSUBTEX				2
#define NCOLS				2
#define DENT_VERSION		1

#define PB_SIZE				0
#define PB_KM				1
#define PB_NITS				2
#define PB_COL1				3
#define PB_COL2				4

#define DENT_VERS			0xDE01
#define MTL_HDR_CHUNK		0x4000
#define DENTVERS1_CHUNK		0x4001
#define DENTSIZE			50.0f
#define NOISE_DIM			20    
#define FNOISE_DIM			20.0


//	C Prototypes
ClassDesc*		GetDentDesc();
static BOOL		CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam);



//	Class Declerations
class Dent;

class DentDlg: public ParamDlg 
{
	public:
		HWND hwmedit;	 	// window handle of the materials editor dialog
		IMtlParams *ip;
		Dent *theTex;	// current Dent being edited.
		HWND hPanel; 		// Rollup pane
		ISpinnerControl *kmSpin,*nitsSpin,*sizeSpin;
		IColorSwatch *cs[2];
		TimeValue curTime; 
		ParamDlg *xyzGenDlg;
		int isActive;
		//-----------------------------
		DentDlg(HWND hwMtlEdit, IMtlParams *imp, Dent *m); 
		~DentDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void UpdateSubTexNames();
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff);

		// methods inherited from ParamDlg:
		Class_ID ClassID();
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};

class Dent: public Tex3D 
{ 
	friend class DentPostLoad;
	friend class DentDlg;
	Color col[NCOLS];
	float km,size;
	int	nits;
	XYZGen *xyzGen;		   // ref #0
	IParamBlock *pblock;   // ref #1
	Texmap* subTex[NSUBTEX];  // More refs
	Interval ivalid;
	int rollScroll;
	int vers;
	DentDlg *paramDlg;
	float DentFunc(float x, float y, float z);

	public:
		Dent();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }

		void SetColor(int i, Color c, TimeValue t);
		void SetSize(float f, TimeValue t);
		void SetKm(float f, TimeValue t);
		void SetNits(int f, TimeValue t);
		void NotifyChanged();
		void SwapInputs(); 

		void ReadSXPData(TCHAR *name, void *sxpdata);

		// Dent noise functions
		void	InitNoise();
		float	DentNoise(float x, float y, float z);
		void	LerpColor(RGBA *c, RGBA *a, RGBA *b, float f);
		float	SmoothStep(float x0, float x1, float v);

		// Evaluate the color of map for the context.
		RGBA EvalColor(ShadeContext& sc);

		// For Bump mapping, need a perturbation to apply to a normal.
		// Leave it up to the Texmap to determine how to do this.
		Point3 EvalNormalPerturb(ShadeContext& sc);

		// Requirements
		ULONG LocalRequirements(int subMtlNum) { return xyzGen->Requirements(subMtlNum); }

		// Methods to access texture maps of material
		int NumSubTexmaps() { return NSUBTEX; }
		Texmap* GetSubTexmap(int i) { return subTex[i]; }
		void SetSubTexmap(int i, Texmap *m);
		TSTR GetSubTexmapSlotName(int i);

		Class_ID ClassID();
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_DENT); }  
		void DeleteThis() { delete this; }	

		int NumSubs() { return 2+NSUBTEX; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return 2+NSUBTEX; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

};

class DentClassDesc:public ClassDesc 
{
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new Dent; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_DENT); }
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID();
	const TCHAR* 	Category() { return TEXMAP_CAT_3D;  }
};

class DentPostLoad : public PostLoadCallback 
{
	public:
		Dent *chk;
		DentPostLoad(Dent *b) {chk=b;}
		void proc(ILoad *iload) {
			if (chk->vers<1) {
				if (chk->pblock) 
					ScaleFloatController(chk->pblock, PB_SIZE, 100.0f);
//				iload->SetObsolete();
				}
			delete this;
			}
};


#pragma pack(1)
struct DentState 
{
	ulong	version;
	float	size, km;
	Col24	col1, col2;
	int		nits;
};
#pragma pack()

#endif
