/**********************************************************************
 *<
	FILE: MIRROR.CPP

	DESCRIPTION: Automatic cubic map.

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include <bmmlib.h>
#include "render.h"
#include "stdmat.h"

//#define DBG
extern HINSTANCE hInstance;

static Class_ID mirrorClassID(MIRROR_CLASS_ID,0);


#define PB_BLUR	    0

//---------------------------------------------------

class MirrorMap {	
	public:
		Bitmap  *bm;
		IPoint2 org;
		int devW;
		int nodeID;
		MirrorMap *next;
		MirrorMap() { next = NULL; bm = NULL; nodeID = -1; }
		~MirrorMap() { FreeMap(); }
		void FreeMap() { if (bm) delete bm; bm = NULL; } 
		
		int AllocMap(int w, int h);
	};

int MirrorMap::AllocMap(int w, int h) {
	if ( bm && w==bm->Width() && h==bm->Height())
		return 1;
	BitmapInfo bi;
	if (bm) delete bm;
	bi.SetName(_T(""));
	bi.SetWidth(w);
	bi.SetHeight(h);
	bi.SetType(BMM_TRUE_32);
	bm = TheManager->Create(&bi);
	return 1;
	}


class Mirror;

class MirrorDlg: public ParamDlg {
	public:
		HWND hwmedit;	 	// window handle of the materials editor dialog
		IMtlParams *ip;
		Mirror *theTex;	// current Mirror being edited.
		HWND hPanel; 		// Rollup pane
		ISpinnerControl *blurSpin,*nthSpin;
		TimeValue curTime; 
		int isActive;
		BOOL valid;

		//-----------------------------
		MirrorDlg(HWND hwMtlEdit, IMtlParams *imp, Mirror *m); 
		~MirrorDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff);
		void Invalidate() { valid = FALSE;	InvalidateRect(hPanel,NULL,0); }

		// methods inherited from ParamDlg:
		Class_ID ClassID() {return mirrorClassID;  }
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};


//--------------------------------------------------------------
// Mirror: 
//--------------------------------------------------------------

class Mirror: public StdMirror { 
	friend class MirrorPostLoad;
	friend class MirrorDlg;
	int nth;
	float blur;
	BOOL applyBlur;
	BOOL do_nth;
	BOOL useEnvMap;
	IParamBlock *pblock;   // ref #1
    MirrorMap *maps;
	TimeValue mapTime;
	Interval ivalid;
	int rollScroll;
	MirrorDlg *paramDlg;
	public:
		Mirror();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }

		// methods inherited from StdMirror:
		void SetDoNth(BOOL onoff) { do_nth = onoff;}
		void SetNth(int n){ nth = n;}
		void SetApplyBlur(BOOL onoff) { applyBlur = onoff; }
		void SetBlur(float b, TimeValue t);
		BOOL GetDoNth() { return do_nth; }
		int GetNth() { return nth;}
		BOOL GetApplyBlur() { return applyBlur;}
		float GetBlur(TimeValue t) { return pblock->GetFloat(PB_BLUR,t); }


		void NotifyChanged();
	
		// Evaluate the color of map for the context.
		RGBA EvalColor(ShadeContext& sc);

		// optimized evaluation for monochrome use
		float EvalMono(ShadeContext& sc);

		// For Bump mapping, need a perturbation to apply to a normal.
		// Leave it up to the Texmap to determine how to do this.
		Point3 EvalNormalPerturb(ShadeContext& sc);

		ULONG LocalRequirements(int subMtlNum) {	
			return MTLREQ_AUTOMIRROR;
			}

		int BuildMaps(TimeValue t, RenderMapsContext &rmc);
		int DoThisFrame(TimeValue t, BOOL fieldRender);
		MirrorMap *FindMap(int nodeNum);
		void FreeMaps();
		int RenderBegin(TimeValue t, ULONG flags) { mapTime = -999999999; return 1;}
		int RenderEnd(TimeValue t) { FreeMaps(); return 1; }

		Class_ID ClassID() {	return mirrorClassID; }
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_FLATMIRROR); }  
		void DeleteThis() { delete this; }	

		int NumSubs() {return 1; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return 1; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

	};

class MirrorClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new Mirror; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_FLATMIRROR); }  
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID() { return mirrorClassID; }
	const TCHAR* 	Category() { return TEXMAP_CAT_2D;  }
	};

static MirrorClassDesc mirrorCD;

ClassDesc* GetMirrorDesc() { return &mirrorCD;  }

static BOOL CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	MirrorDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (MirrorDlg*)lParam;
		theDlg->hPanel = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (MirrorDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = 1;
	int	res = theDlg->PanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = 0;
	return res;
	}

MirrorDlg::MirrorDlg(HWND hwMtlEdit, IMtlParams *imp, Mirror *m) { 
	hwmedit = hwMtlEdit;
	ip = imp;
	hPanel = NULL;
	theTex = m; 
	isActive = 0;
	valid = FALSE;
	hPanel = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_MIRROR),
		PanelDlgProc, 
		GetString(IDS_DS_MIRROR_PARAMS), 
		(LPARAM)this );		
	curTime = imp->GetTime();
	}

void MirrorDlg::ReloadDialog() {
	Interval valid;
	theTex->Update(curTime, valid);
	LoadDialog(FALSE);
	}

void MirrorDlg::SetTime(TimeValue t) {
	Interval valid;
	if (t!=curTime) {
		curTime = t;
		theTex->Update(curTime, valid);
		LoadDialog(FALSE);
		InvalidateRect(hPanel,NULL,0);
		}
	}

MirrorDlg::~MirrorDlg() {
	theTex->paramDlg = NULL;
	ReleaseISpinner(blurSpin);
	ReleaseISpinner(nthSpin);
	SetWindowLong(hPanel, GWL_USERDATA, NULL);
	hPanel =  NULL;
	}

BOOL MirrorDlg::PanelProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			blurSpin = SetupFloatSpinner(hwndDlg, IDC_ACUBE_BLUR_SPIN, IDC_ACUBE_BLUR_EDIT,0.0f,100.0f,1.0f,.01f);
			nthSpin = SetupIntSpinner(hwndDlg, IDC_ACUBE_NTH_SPIN, IDC_ACUBE_NTH_EDIT,1,1000, 1);
			CheckRadioButton( hwndDlg, IDC_FIRST_ONLY, IDC_EVERY_NTH, IDC_FIRST_ONLY+theTex->do_nth);
			SetCheckBox(hwndDlg, IDC_ACUBE_BLUR, theTex->applyBlur);
			SetCheckBox(hwndDlg, IDC_USE_ENVMAP, theTex->useEnvMap);
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_FIRST_ONLY:
					theTex->do_nth = FALSE;
					break;
				case IDC_EVERY_NTH:
					theTex->do_nth = TRUE;
					break;
				case IDC_ACUBE_BLUR:
					theTex->applyBlur = GetCheckBox(hwndDlg, IDC_ACUBE_BLUR);			
					break;
				case IDC_USE_ENVMAP:
					theTex->useEnvMap = GetCheckBox(hwndDlg, id);			
					break;
				}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case WM_PAINT: 	
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			break;
		case WM_CLOSE: 	break;       
		case WM_DESTROY:		break;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (id) {
				case IDC_ACUBE_BLUR_SPIN: 
					theTex->SetBlur(blurSpin->GetFVal(),curTime); 	
					break;
				case IDC_ACUBE_NTH_SPIN: 
					theTex->nth = nthSpin->GetIVal(); 	
					break;
				}
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			theTex->NotifyChanged();
		    UpdateMtlDisplay();
			break;
    	}
	return FALSE;
	}
 

void MirrorDlg::LoadDialog(BOOL draw) {
	if (theTex) {
		Interval valid;
		theTex->Update(curTime,valid);
		blurSpin->SetValue(theTex->blur,FALSE);
		nthSpin->SetValue(theTex->nth,FALSE);
		CheckRadioButton( hPanel, IDC_FIRST_ONLY, IDC_EVERY_NTH, IDC_FIRST_ONLY+theTex->do_nth);
		SetCheckBox(hPanel, IDC_ACUBE_BLUR, theTex->applyBlur);
		}
	}

void MirrorDlg::SetThing(ReferenceTarget *m) {
	assert (m->ClassID()==mirrorClassID);
	assert (m->SuperClassID()==TEXMAP_CLASS_ID);
	if (theTex) theTex->paramDlg = NULL;
	theTex = (Mirror *)m;
	if (theTex)
		theTex->paramDlg = this;
	LoadDialog(TRUE);
	}

void MirrorDlg::ActivateDlg(BOOL onOff) {
	}

//-----------------------------------------------------------------------------
//  Mirror
//-----------------------------------------------------------------------------


#define NPARAMS 1
#define MIRROR_VERSION 2

// Version 1 desc
static ParamBlockDescID pbdesc1[] = {
	{ TYPE_INT, NULL, TRUE,0 }, 	// blur
	{ TYPE_FLOAT, NULL, TRUE,0 } 	// blurOff
	};

// Version 2 desc
static ParamBlockDescID pbdesc2[] = {
	{ TYPE_FLOAT, NULL, TRUE,1 } 	// blur
	};


void Mirror::Reset() {
	ReplaceReference( 0, CreateParameterBlock( pbdesc2, NPARAMS, MIRROR_VERSION) );	
	ivalid.SetEmpty();
	SetBlur(1.0f, TimeValue(0));
	}

void Mirror::NotifyChanged() {
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	}

Mirror::Mirror() {
	paramDlg = NULL;
	pblock = NULL;
	nth = 1;
	do_nth = TRUE;
	applyBlur = TRUE;
	useEnvMap = TRUE;
	maps = NULL;
	Reset();
	rollScroll=0;
	}

RefTargetHandle Mirror::Clone(RemapDir &remap) {
	Mirror *mnew = new Mirror();
	*((MtlBase*)mnew) = *((MtlBase*)this);  // copy superclass stuff
	mnew->ReplaceReference(0,remap.CloneRef(pblock));
	mnew->do_nth = do_nth;
	mnew->applyBlur = applyBlur;
	mnew->ivalid.SetEmpty();	
	return (RefTargetHandle)mnew;
	}

ParamDlg* Mirror::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp) {
	MirrorDlg *dm = new MirrorDlg(hwMtlEdit, imp, this);
	dm->LoadDialog(TRUE);	
	paramDlg = dm;
	return dm;	
	}

void Mirror::Update(TimeValue t, Interval& valid) {		
	if (!ivalid.InInterval(t)) {
		ivalid.SetInfinite();
		pblock->GetValue( PB_BLUR, t, blur, ivalid );
		}
	valid &= ivalid;
	}

void Mirror::FreeMaps() {
	MirrorMap *cm,*nxtcm;
	for (cm = maps; cm!=NULL; cm = nxtcm) {
		nxtcm = cm->next;
	   	delete cm;		
		}
	maps = NULL;
	}


void Mirror::SetBlur(float f, TimeValue t) { 
	blur = f; 
	pblock->SetValue( PB_BLUR, t, f);
	}


RefTargetHandle Mirror::GetReference(int i) {
	switch(i) {
		case 0:	return pblock ;
		default: return NULL;
		}
	}

void Mirror::SetReference(int i, RefTargetHandle rtarg) {
	switch(i) {
		case 0:	pblock = (IParamBlock *)rtarg; break;
		}
	}
	 
Animatable* Mirror::SubAnim(int i) {
	switch (i) {
		case 0: return pblock;
		default: return NULL;
		}
	}

TSTR Mirror::SubAnimName(int i) {
	switch (i) {
		case 0: return TSTR(GetString(IDS_DS_PARAMETERS));
		default: return TSTR("");		
		}
	}

static int nameID[] = { IDS_DS_BLUR };

RefResult Mirror::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
   PartID& partID, RefMessage message ) {
	switch (message) {
		case REFMSG_CHANGE:
			if (paramDlg&&!paramDlg->isActive) 
				paramDlg->Invalidate();
			break;

		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_BLUR: 
					gpd->dim = defaultDim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			gpn->name = TSTR(GetString(nameID[gpn->index]));
			return REF_STOP; 
			}
		}
	return(REF_SUCCEED);
	}

inline AColor AColorFrom64(BMM_Color_64 pix) {
	return AColor((float)pix.r/65536.0f,(float)pix.g/65536.0f,(float)pix.b/65536.0f,(float)pix.a/65536.0f);
	}


static void FlipAxis(Matrix3& tm, int k) {
	MRow* m = tm.GetAddr();
	for (int i=0; i<4; i++) m[i][k] = -m[i][k];
	}

/* build reflection matrix for plane p */
static void BuildReflMatrix(Matrix3& rm, float *p) {
	MRow* m = rm.GetAddr();
	m[0][0] = 1.0f-2.0f*p[0]*p[0];		
	m[1][1] = 1.0f-2.0f*p[1]*p[1];		
	m[2][2] = 1.0f-2.0f*p[2]*p[2];		
	m[0][1] = m[1][0] = -2.0f*p[0]*p[1];		
	m[0][2] = m[2][0] = -2.0f*p[0]*p[2];		
	m[1][2] = m[2][1] = -2.0f*p[1]*p[2];		
	m[3][0] = -2.0f*p[0]*p[3];		
	m[3][1] = -2.0f*p[1]*p[3];		
	m[3][2] = -2.0f*p[2]*p[3];		
	rm.SetNotIdent();
	}


MirrorMap *Mirror::FindMap(int nodeNum) {
	MirrorMap *cm;
	for (cm=maps; cm!=NULL; cm = cm->next)
		if (cm->nodeID==nodeNum) return cm;
	return NULL;
	}

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

int Mirror::DoThisFrame(TimeValue t, BOOL fieldRender) {
	if (t==mapTime) return 1;  // must be another object with this map on the same frame.
	if (!do_nth) return 0;  // only do it once.
	if (nth==1) return 1;   // need every one
	TimeValue del = abs(t - mapTime);
	if (fieldRender) del*=2;
	return  (del>=SecToTicks(nth))?1:0;
	}

int Mirror::BuildMaps(TimeValue t, RenderMapsContext &rmc) {
	SubRendParams srp;
	rmc.GetSubRendParams(srp);
  	MirrorMap *mir = FindMap(rmc.NodeRenderID());
	if (mir&&!DoThisFrame(t,srp.fieldRender))
		return 1;
	mapTime = t;
	ViewParams vp;

	float plane[4];
	Box2 sbox;
    rmc.FindMtlPlane(plane);
	rmc.GetCurrentViewParams(vp);
	Matrix3 tm = Inverse(vp.affineTM);  // get tm = camToWorld
	if (vp.projType==PROJ_PERSPECTIVE) {
		// If this is a perspective projection, and  the Camera ist behind plane
		// of the mirror, then quit.
		if ( DotProd(tm.GetRow(3),Point3(plane[0],plane[1],plane[2]))+plane[3]<0.0f ) 
			return 0;
		}
	else  {
		// If this is an ortho projection, and if view vector is
		// looking at edge or back of mirror, we punt.
		if (DotProd(tm.GetRow(2),Point3(plane[0],plane[1],plane[2]))<.0001f )
			return 0; 
		}

	Matrix3 rm;
	// compute a matrix that reflects in plane
	BuildReflMatrix(rm, plane);
	vp.affineTM  = Inverse(tm*rm);

	// Flip so parity is positive -- otherwise it wont render right. This
	// will be corrected for when we access the bitmap.
	FlipAxis(vp.affineTM, X_AXIS);	

	rmc.FindMtlScreenBox(sbox, &vp.affineTM, rmc.SubMtlIndex());

	int xmin,xmax,ymin,ymax;
	xmax = sbox.right;
	xmin = sbox.left;
	ymax = sbox.bottom;
	ymin = sbox.top;
	if (srp.rendType==RENDTYPE_REGION) {
		int rxmax = srp.devWidth - srp.xmin;
		int rxmin = srp.devWidth - srp.xmax;
		ymin = MAX(ymin,srp.ymin);
		ymax = MIN(ymax,srp.ymax);
		xmin = MAX(xmin,rxmin);
		xmax = MIN(xmax,rxmax);
		}

	if (srp.fieldRender) {
		ymin *= 2;
		ymax *= 2;
		}
	srp.ymin = MAX(ymin,0);
	srp.ymax = MIN(ymax,srp.devHeight);
	srp.xmin = MAX(xmin,0);
	srp.xmax = MIN(xmax,srp.devWidth);
	srp.xorg = srp.xmin;
	srp.yorg = srp.ymin;
	srp.doingMirror = TRUE;
	srp.doEnvMap = useEnvMap;

	int w = srp.xmax-srp.xmin;
	int h = srp.ymax-srp.ymin;

	if (w<=0||h<=0) 
		return 0;
#ifdef DBG
	w = ((w+3)/4)*4;   // For some reason this needs to be a multiple of 4 for bm->Display
#endif

  	if (mir==NULL) {
	  	mir = new MirrorMap;
		mir->nodeID = rmc.NodeRenderID();
		mir->next = maps;
		maps = mir;
		}
	
	mir->AllocMap(w, h);
	mir->org.x = srp.xorg;
	mir->org.y = srp.yorg;
//	mir->org.y = srp.fieldRender ?srp.yorg/2:srp.yorg;
	mir->devW = srp.devWidth;  // TBD: This should be accessable from the SC

	if (!rmc.Render(mir->bm, vp, srp, plane))
		return 0;

	if (srp.fieldRender&&applyBlur) {
		// Double the lines, otherwise the blur wont work right,
		// because the blank lines in between get averaged in and darken it.
		int	evenLines = srp.evenLines; 

		PixelBuf l64(w);
		if (evenLines) {
			for (int i=0; i<h; i+=2) {
				BMM_Color_64 *p64=l64.Ptr();
				if (i+1<h) {
					mir->bm->GetPixels(0,i,  w, p64); 
					mir->bm->PutPixels(0,i+1,w, p64);				
					}
				}
			}
		else {
			for (int i=0; i<h; i+=2) {
				BMM_Color_64 *p64=l64.Ptr();
				if (i+1<h) {
					mir->bm->GetPixels(0,i+1,w, p64); 
					mir->bm->PutPixels(0,i  ,w, p64);				
					}
				}
			}
		}

#ifdef DBG
	mir->bm->UnDisplay();
	mir->bm->Display(_T("Mirror Test"), BMM_UR);
	MessageBox(NULL, _T("hi"), _T(" Mirror Test"), MB_OK|MB_ICONEXCLAMATION);
#endif

	if (applyBlur) {
		// I tried pyramids here, but SATs looked much better.
		mir->bm->SetFilter(BMM_FILTER_SUM); 
		BitmapFilter *filt = mir->bm->Filter();
		if (filt)
			filt->MakeDirty();  // so filter gets recomputed for each frame
		}
	else 
		mir->bm->SetFilter(BMM_FILTER_NONE); 
	return 1;
	}

static BMM_Color_64 black64 = {0,0,0,0};

RGBA Mirror::EvalColor(ShadeContext& sc) {
	BMM_Color_64 c;
	IPoint2 s;
	Point3 V = sc.VectorTo(sc.V(),REF_WORLD);
	int id = sc.NodeID();
	MirrorMap *mir = FindMap(id);
	if (gbufID) sc.SetGBufferID(gbufID);
	if (mir) {
		if (applyBlur) {
			s = sc.ScreenCoord();
			int w = mir->bm->Width(); 
			int h = mir->bm->Height();
			float du=1.0f/float(w);
			float dv=1.0f/float(h);
			float du2= du*blur;
			float dv2= dv*blur;
			mir->bm->GetFiltered(
				float(mir->devW-1-s.x-mir->org.x)*du, 
				float(s.y-mir->org.y)*dv,du2,dv2,&c);
			}
		else {
			int w = mir->bm->Width(); 
			s = sc.ScreenCoord();
			mir->bm->GetPixels(mir->devW-1-s.x-mir->org.x, s.y-mir->org.y,1,&c);
#if 0
			FILE *f = sc.DebugFile();
			if (f) {
				fprintf(f,"w=%d, h=%d org = (%d, %d) ", w, mir->bm->Height(),mir->org.x, mir->org.y);
				fprintf(f, "s = (%d, %d), c = (%x, %x, %x, %x)\n",
					s.x,s.y,c.r,c.g,c.b,c.a);
				}
#endif
			}
		return AColorFrom64(c);
		}
	else 
		return Color(0,0,0);
	}

float Mirror::EvalMono(ShadeContext& sc) {
	return Intens(EvalColor(sc));
	}

Point3 Mirror::EvalNormalPerturb(ShadeContext& sc) {
	return Point3(0,0,0);
	}

#define MTL_HDR_CHUNK 0x4000
#define DONT_DO_NTH_CHUNK 0x1000
#define NTH_CHUNK 0x1001
#define DONT_APPLY_BLUR_CHUNK 0x1002
#define DONT_USE_ENV_CHUNK 0x1003

IOResult Mirror::Save(ISave *isave) { 
	IOResult res;
	ULONG nb;
	// Save common stuff
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();
	if (!do_nth) {
		isave->BeginChunk(DONT_DO_NTH_CHUNK);
		isave->EndChunk();
		}
	if (!applyBlur) {
		isave->BeginChunk(DONT_APPLY_BLUR_CHUNK);
		isave->EndChunk();
		}
	if (!useEnvMap) {
		isave->BeginChunk(DONT_USE_ENV_CHUNK);
		isave->EndChunk();
		}
	isave->BeginChunk(NTH_CHUNK);
	isave->Write(&nth,sizeof(nth),&nb);			
	isave->EndChunk();
	return IO_OK;
	}	

class MirrorPostLoad : public PostLoadCallback {
	public:
		Mirror *theMir;
		MirrorPostLoad(Mirror *b) {theMir=b;}
		void proc(ILoad *iload) {
			if (theMir->pblock->GetVersion()!=MIRROR_VERSION) {
				switch (theMir->pblock->GetVersion()) {
					case 1:
						theMir->ReplaceReference(0,
							UpdateParameterBlock(
								pbdesc1, 2, theMir->pblock,
								pbdesc2, 1, MIRROR_VERSION));
						iload->SetObsolete();
						break;
					default:
						assert(0);
						break;
					}
				}
			//waitPostLoad--;
			delete this;
			}
	};


IOResult Mirror::Load(ILoad *iload) { 
	ULONG nb;
	IOResult res;
	iload->RegisterPostLoadCallback(new MirrorPostLoad(this));
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			case DONT_DO_NTH_CHUNK:
				do_nth = FALSE;
				break;
			case DONT_APPLY_BLUR_CHUNK:
				applyBlur = FALSE;
				break;
			case DONT_USE_ENV_CHUNK:
				useEnvMap = FALSE;
				break;
			case NTH_CHUNK:
				iload->Read(&nth,sizeof(nth),&nb);			
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	return IO_OK;
	}
