/**********************************************************************
 *<
	FILE: MIX.CPP

	DESCRIPTION: MIX Composite.

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include "stdmat.h"

extern HINSTANCE hInstance;

static LRESULT CALLBACK CurveWndProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam );


#define NSUBTEX 3    // number of texture map slots
#define NCOLS 2      // number of color swatches

static Class_ID mixClassID(MIX_CLASS_ID+10,9999);


class Mix;

class MixDlg: public ParamDlg {
	public:
		HWND hwmedit;	 // window handle of the materials editor dialog
		HWND hwCurve;	 // window handle of the mix curve window
		IMtlParams *ip;
		Mix *theTex;	 // current Mix being edited.
		HWND hPanel; // Rollup pane
		ISpinnerControl *mixSpin;
		ISpinnerControl *crvASpin;
		ISpinnerControl *crvBSpin;
		IColorSwatch *cs[NCOLS];
		TimeValue curTime; 
		BOOL valid;
		BOOL isActive;

		//-----------------------------
		MixDlg(HWND hwMtlEdit, IMtlParams *imp, Mix *m); 
		~MixDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		BOOL CurveWindowProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam );
		void DrawCurve(HDC hdc, Rect& rect);
		void UpdateCurve();

		void UpdateSubTexNames();
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff);
		void EnableControls();
		void Invalidate() { valid = FALSE;	InvalidateRect(hPanel,NULL,0); }

		// methods inherited from ParamDlg:
		Class_ID ClassID() {return mixClassID;  }
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};



//--------------------------------------------------------------
// Mix: A Composite texture map
//--------------------------------------------------------------
class Mix: public MultiTex { 
	friend class MixPostLoad;
	friend class MixDlg;
	MixDlg *paramDlg;
	Color col[NCOLS];
	IParamBlock *pblock;   // ref #0
	Texmap* subTex[NSUBTEX];  // More refs
	Interval ivalid;
	float mix;
	BOOL useCurve;
	float crvA;
	float crvB;
	BOOL rollScroll;
	public:
		Mix();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }

		void SetColor(int i, Color c, TimeValue t);
		void SetMix(float f, TimeValue t);
		void SetCrvA(float f, TimeValue t);
		void SetCrvB(float f, TimeValue t);
		void NotifyChanged();
		void SwapInputs(); 

		float mixCurve(float x);

		// Evaluate the color of map for the context.
		AColor EvalColor(ShadeContext& sc);
		float EvalMono(ShadeContext& sc);
		AColor EvalFunction(ShadeContext& sc, float u, float v, float du, float dv);

		// For Bump mapping, need a perturbation to apply to a normal.
		// Leave it up to the Texmap to determine how to do this.
		Point3 EvalNormalPerturb(ShadeContext& sc);

		// Methods to access texture maps of material
		int NumSubTexmaps() { return NSUBTEX; }
		Texmap* GetSubTexmap(int i) { return subTex[i]; }
		void SetSubTexmap(int i, Texmap *m);
		TSTR GetSubTexmapSlotName(int i);

		Class_ID ClassID() {	return mixClassID; }
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_MIX); }  
		void DeleteThis() { delete this; }	

		int NumSubs() { return 1+NSUBTEX; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return 1+NSUBTEX; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

	};

int numMixs = 0;
class MixClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new Mix; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_MIX); }
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID() { return mixClassID; }
	const TCHAR* 	Category() { return TEXMAP_CAT_COMP;  }
	};

static MixClassDesc mixCD;

ClassDesc* GetMixDesc() { return &mixCD;  }

static BOOL CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	MixDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (MixDlg*)lParam;
		theDlg->hPanel = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (MixDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = TRUE;
	int	res = theDlg->PanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = FALSE;
	return res;
	}

MixDlg::MixDlg(HWND hwMtlEdit, IMtlParams *imp, Mix *m) { 
	hwmedit = hwMtlEdit;
	ip = imp;
	hPanel = NULL;
	theTex = m; 
	valid = FALSE;
	isActive = FALSE;
	hPanel = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_MIX),
		PanelDlgProc, 
		GetString(IDS_DS_MIXPARMS), 
		(LPARAM)this );		
	curTime = imp->GetTime();
	}

void MixDlg::ReloadDialog() {
	Interval valid;
	theTex->Update(curTime, valid);
	LoadDialog(FALSE);
	}

void MixDlg::SetTime(TimeValue t) {
	Interval valid;
	if (t!=curTime) {
		curTime = t;
		theTex->Update(curTime, valid);
		LoadDialog(FALSE);
		InvalidateRect(hPanel,NULL,0);
		}
	}

MixDlg::~MixDlg() {
	theTex->paramDlg = NULL;
	ReleaseISpinner(mixSpin);
	ReleaseISpinner(crvASpin);
	ReleaseISpinner(crvBSpin);
	ReleaseIColorSwatch(cs[0]);
	ReleaseIColorSwatch(cs[1]);
	SetWindowLong(hPanel, GWL_USERDATA, NULL);
	hPanel =  NULL;
	}

static int colID[2] = { IDC_MIX_COL1, IDC_MIX_COL2 };

BOOL MixDlg::PanelProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			mixSpin = SetupFloatSpinner(hwndDlg, IDC_MIX_SPIN, IDC_MIX_EDIT,0.0f,1.0f,.01f);
			mixSpin->SetScale( 0.01f );
			crvASpin = SetupFloatSpinner(hwndDlg, IDC_MIXA_SPIN, IDC_MIXA_EDIT,0.0f,1.0f,.25f,.01f);
			crvBSpin = SetupFloatSpinner(hwndDlg, IDC_MIXB_SPIN, IDC_MIXB_EDIT,0.0f,1.0f,.75f,.01f);
			for (int i=0; i<NCOLS; i++) 
   				cs[i] = GetIColorSwatch(GetDlgItem(hwndDlg, colID[i]),
   					theTex->col[i],theTex->GetSubTexmapSlotName(i).data());
			hwCurve = GetDlgItem(hwndDlg, IDC_MIXCURVE);
			SetWindowLong( hwCurve, GWL_WNDPROC, (LONG)CurveWndProc);
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_MIX_TEX1: 	PostMessage(hwmedit,WM_TEXMAP_BUTTON,0 ,(LPARAM)theTex);	break;
				case IDC_MIX_TEX2: 	PostMessage(hwmedit,WM_TEXMAP_BUTTON,1 ,(LPARAM)theTex);	break;
				case IDC_MIX_TEX3: 	PostMessage(hwmedit,WM_TEXMAP_BUTTON,2 ,(LPARAM)theTex);	break;
				case IDC_MIX_USECURVE: 
					theTex->useCurve =  GetCheckBox(hwndDlg, id);			
					UpdateMtlDisplay();
					theTex->NotifyChanged();
					break;
				case IDC_MIX_SWAP:
					theTex->SwapInputs(); 
					cs[0]->SetColor(theTex->col[0]);
					cs[1]->SetColor(theTex->col[1]);
					UpdateSubTexNames();
					UpdateMtlDisplay();
					theTex->NotifyChanged();
					break;
				}
			break;
		case CC_COLOR_BUTTONDOWN:
			theHold.Begin();
			break;
		case CC_COLOR_BUTTONUP:
			if (HIWORD(wParam)) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			break;
		case CC_COLOR_CHANGE:
			{
			int id = LOWORD(wParam);
			int buttonUp = HIWORD(wParam); 
			int n = (id==IDC_MIX_COL1)?0:1;
			if (buttonUp) theHold.Begin();
			theTex->SetColor(n,cs[n]->GetColor(),curTime);
			if (buttonUp) {
				theHold.Accept(GetString(IDS_DS_PARAMCHG));
				theTex->NotifyChanged();
				}
			}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case WM_PAINT: 	
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			break;
		case WM_CLOSE: 	break;       
		case WM_DESTROY:  	break;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (id) {
				case IDC_MIX_SPIN: 	theTex->SetMix(mixSpin->GetFVal(),curTime);  break;
				case IDC_MIXA_SPIN: theTex->SetCrvA(crvASpin->GetFVal(),curTime);  break;
				case IDC_MIXB_SPIN: theTex->SetCrvB(crvBSpin->GetFVal(),curTime);  break;
				}
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			theTex->NotifyChanged();
		    UpdateMtlDisplay();
			break;

    	}
	return FALSE;
	}


static int subTexId[NSUBTEX] = { IDC_MIX_TEX1, IDC_MIX_TEX2, IDC_MIX_TEX3 };

void MixDlg::UpdateSubTexNames() {
	for (int i=0; i<NSUBTEX; i++) {
		Texmap *m = theTex->subTex[i];
		TSTR nm;
		if (m) 	nm = m->GetFullName();
		else 	nm = GetString(IDS_DS_NONE);
		SetDlgItemText(hPanel, subTexId[i], nm.data());
		}
	}

void MixDlg::LoadDialog(BOOL draw) {
	if (theTex) {
		Interval valid;
		theTex->Update(curTime,valid);
		mixSpin->SetValue(theTex->mix,FALSE);
		crvASpin->SetValue(theTex->crvA,FALSE);
		crvBSpin->SetValue(theTex->crvB,FALSE);
		cs[0]->SetColor(theTex->col[0]);
		cs[1]->SetColor(theTex->col[1]);
		SetCheckBox(hPanel, IDC_MIX_USECURVE, theTex->useCurve);
		UpdateSubTexNames();
		UpdateCurve();
		EnableControls();
		}
	}

void MixDlg::SetThing(ReferenceTarget *m) {
	assert (m->ClassID()==mixClassID);
	assert (m->SuperClassID()==TEXMAP_CLASS_ID);
	if (theTex) theTex->paramDlg = NULL;
	theTex = (Mix *)m;
	if (theTex) theTex->paramDlg = this;
	LoadDialog(TRUE);
	}

void MixDlg::ActivateDlg(BOOL onOff) {
	for (int i=0; i<NCOLS; i++)
		cs[i]->Activate(onOff);
	}

void MixDlg::EnableControls() {
	if (hPanel==NULL) return;
	for (int i=0; i<2; i++)
		EnableWindow(GetDlgItem(hPanel, colID[i]), theTex->subTex[i]?0:1);
	BOOL b = theTex->subTex[2]?0:1;
	EnableWindow(GetDlgItem(hPanel,IDC_MIX_SPIN),b);
	EnableWindow(GetDlgItem(hPanel,IDC_MIX_EDIT),b);
	}


//-----------------------------------------------------------------------------
//  Mix
//-----------------------------------------------------------------------------

#define MIX_VERSION 2

#define PB_COL1	0
#define PB_COL2	1
#define PB_MIX	2
#define PB_CRVA	3
#define PB_CRVB	4

#define NPARAMS 5

static int name_id[NPARAMS] = {IDS_DS_COLOR1, IDS_DS_COLOR2, IDS_DS_MIXAMT,
	IDS_RB_LOWER,IDS_RB_UPPER };

static ParamBlockDescID pbdesc1[] = {
	{ TYPE_RGBA, NULL, TRUE,1 },   // col1
	{ TYPE_RGBA, NULL, TRUE,2 },   // col2
	{ TYPE_FLOAT, NULL, TRUE,3} 	 // mix
	};   

static ParamVersionDesc oldVersions[1] = {
	ParamVersionDesc(pbdesc1,3,1)
	};

static ParamBlockDescID pbdesc2[] = {
	{ TYPE_RGBA, NULL, TRUE,1 },   // col1
	{ TYPE_RGBA, NULL, TRUE,2 },   // col2
	{ TYPE_FLOAT, NULL, TRUE,3}, 	 // mix
	{ TYPE_FLOAT, NULL, TRUE,4}, 	 // crva
	{ TYPE_FLOAT, NULL, TRUE,5} 	 // crvb
	};   
static ParamVersionDesc curVersion(pbdesc2,5,002);

void Mix::Reset() {
	useCurve = 0;
	ReplaceReference( 0, CreateParameterBlock( pbdesc2, NPARAMS, MIX_VERSION) );	
	ivalid.SetEmpty();
	for (int i=0; i<NSUBTEX; i++) {
		DeleteReference(i+1);	// get rid of maps
		}
	SetColor(0, Color(0.0f,0.0f,0.0f), TimeValue(0));
	SetColor(1, Color(1.0f,1.0f,1.0f), TimeValue(0));
	SetMix(.0f, TimeValue(0));
	SetCrvA(.3f, TimeValue(0));
	SetCrvB(.7f, TimeValue(0));
	}

void Mix::NotifyChanged() {
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	}

Mix::Mix() {
	for (int i=0; i<NSUBTEX; i++) subTex[i] = NULL;
	pblock = NULL;
	paramDlg = NULL;
	Reset();
	rollScroll=0;
	}

float Mix::mixCurve(float x) {
	if (x<crvA)  return 0.0f;
	if (x>=crvB) return 1.0f;
	x = (x-crvA)/(crvB-crvA);
	return (x*x*(3-2*x));
	}

AColor Mix::EvalColor(ShadeContext& sc) {
	if (gbufID) sc.SetGBufferID(gbufID);
	float m = subTex[2]? subTex[2]->EvalMono(sc): mix;
	if (useCurve) m = mixCurve(m);
	if (m==0.0f)
		return subTex[0] ? subTex[0]->EvalColor(sc): col[0];
	else if(m==1.0f)   
		return subTex[1] ? subTex[1]->EvalColor(sc): col[1];
	else {
		AColor c0  = subTex[0] ? subTex[0]->EvalColor(sc): col[0];
		AColor c1  = subTex[1] ? subTex[1]->EvalColor(sc): col[1];
		return m*c1 + (1.0f-m)*c0;
		}
	}

float Mix::EvalMono(ShadeContext& sc) {
	if (gbufID) sc.SetGBufferID(gbufID);
	float m = subTex[2]? subTex[2]->EvalMono(sc): mix;
	if (useCurve) m = mixCurve(m);
	if (m==0.0f)
		return subTex[0] ? subTex[0]->EvalMono(sc): Intens(col[0]);
	else if(m==1.0f)   
		return subTex[1] ? subTex[1]->EvalMono(sc): Intens(col[1]);
	else {
		float c0  = subTex[0] ? subTex[0]->EvalMono(sc): Intens(col[0]);
		float c1  = subTex[1] ? subTex[1]->EvalMono(sc): Intens(col[1]);
		return m*c1 + (1.0f-m)*c0;
		}
	}

Point3 Mix::EvalNormalPerturb(ShadeContext& sc) {
	if (gbufID) sc.SetGBufferID(gbufID);
	float m = subTex[2]? subTex[2]->EvalMono(sc): mix;
	if (useCurve) m = mixCurve(m);
	if (m==0.0f)
		return subTex[0] ? subTex[0]->EvalNormalPerturb(sc): Point3(0,0,0);
	else if(m==1.0f)   
		return subTex[1] ? subTex[1]->EvalNormalPerturb(sc): Point3(0,0,0);
	else {
		Point3 p0  = subTex[0] ? subTex[0]->EvalNormalPerturb(sc): Point3(0,0,0);
		Point3 p1  = subTex[1] ? subTex[1]->EvalNormalPerturb(sc): Point3(0,0,0);
		return m*p1 + (1.0f-m)*p0;
		}
	}

RefTargetHandle Mix::Clone(RemapDir &remap) {
	Mix *mnew = new Mix();
	*((MtlBase*)mnew) = *((MtlBase*)this);  // copy superclass stuff
	mnew->ReplaceReference(0,remap.CloneRef(pblock));
	mnew->col[0] = col[0];
	mnew->col[1] = col[1];
	mnew->mix = mix;
	mnew->crvA = crvA;
	mnew->crvB = crvB;
	mnew->useCurve = useCurve;
	mnew->ivalid.SetEmpty();	
	for (int i = 0; i<NSUBTEX; i++) {
		mnew->subTex[i] = NULL;
		if (subTex[i])
			mnew->ReplaceReference(i+1,remap.CloneRef(subTex[i]));
		}
	return (RefTargetHandle)mnew;
	}

ParamDlg* Mix::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp) {
	MixDlg *dm = new MixDlg(hwMtlEdit, imp, this);
	dm->LoadDialog(TRUE);	
	paramDlg = dm;
	return dm;	
	}


void Mix::Update(TimeValue t, Interval& valid) {		
	if (!ivalid.InInterval(t)) {
		ivalid.SetInfinite();
		pblock->GetValue( PB_COL1, t, col[0], ivalid );
		col[0].ClampMinMax();
		pblock->GetValue( PB_COL2, t, col[1], ivalid );
		col[1].ClampMinMax();
		pblock->GetValue( PB_MIX, t, mix, ivalid );
		pblock->GetValue( PB_CRVA, t, crvA, ivalid );
		pblock->GetValue( PB_CRVB, t, crvB, ivalid );
		for (int i=0; i<NSUBTEX; i++) {
			if (subTex[i]) 
				subTex[i]->Update(t,ivalid);
			}
		}
	valid &= ivalid;
	}


void Mix::SwapInputs() {
	Color t = col[0]; col[0] = col[1]; col[1] = t;
	Texmap *x = subTex[0];  subTex[0] = subTex[1];  subTex[1] = x;
	pblock->SwapControllers(PB_COL1,PB_COL2);
	}

void Mix::SetColor(int i, Color c, TimeValue t) {
    col[i] = c;
	pblock->SetValue( i==0?PB_COL1:PB_COL2, t, c);
	}

void Mix::SetMix(float f, TimeValue t) { 
	mix = f; 
	pblock->SetValue( PB_MIX, t, f);
	}

void Mix::SetCrvA(float f, TimeValue t) { 
	crvA = f; 
	pblock->SetValue( PB_CRVA, t, f);
	if (crvB<crvA) {
		crvB = crvA;
		pblock->SetValue( PB_CRVB, t, crvB);
		if (paramDlg)
			paramDlg->crvBSpin->SetValue(crvB,FALSE);
		}

	if (paramDlg)
		paramDlg->UpdateCurve();
	}

void Mix::SetCrvB(float f, TimeValue t) { 
	crvB = f; 
	pblock->SetValue( PB_CRVB, t, f);
	if (crvB<crvA) {
		crvA = crvB;
		pblock->SetValue( PB_CRVA, t, crvA);
		if (paramDlg)
			paramDlg->crvASpin->SetValue(crvA,FALSE);
		}
	if (paramDlg)
		paramDlg->UpdateCurve();
	}

RefTargetHandle Mix::GetReference(int i) {
	switch(i) {
		case 0:	return pblock ;
		default:return subTex[i-1];
		}
	}

void Mix::SetReference(int i, RefTargetHandle rtarg) {
	switch(i) {
		case 0:	pblock = (IParamBlock *)rtarg; break;
		default: subTex[i-1] = (Texmap *)rtarg; break;
		}
	}

void Mix::SetSubTexmap(int i, Texmap *m) {
	ReplaceReference(i+1,m);
	if (paramDlg) {
		paramDlg->UpdateSubTexNames();
		paramDlg->EnableControls();
		}
	}

TSTR Mix::GetSubTexmapSlotName(int i) {
	switch(i) {
		case 0:  return TSTR(GetString(IDS_DS_COLOR1)); 
		case 1:  return TSTR(GetString(IDS_DS_COLOR2)); 
		case 2:  return TSTR(GetString(IDS_DS_MIXAMT)); 
		default: return TSTR(_T(""));
		}
	}
	 
Animatable* Mix::SubAnim(int i) {
	switch (i) {
		case 0: return pblock;
		default: return subTex[i-1]; 
		}
	}

TSTR Mix::SubAnimName(int i) {
	switch (i) {
		case 0: return TSTR(GetString(IDS_DS_PARAMETERS));		
		default: return GetSubTexmapTVName(i-1);
		}
	}

RefResult Mix::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
   PartID& partID, RefMessage message ) {
	switch (message) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			if (paramDlg&&!paramDlg->isActive) 
				paramDlg->Invalidate();
			break;

		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_COL1: 
				case PB_COL2: gpd->dim = stdColor255Dim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			gpn->name= TSTR(GetString(name_id[gpn->index]));
			return REF_STOP; 
			}
		}
	return(REF_SUCCEED);
	}


#define MTL_HDR_CHUNK 0x4000
#define USE_CURVE_CHUNK 0x6000
#define VERS2_CHUNK 0x6001

IOResult Mix::Save(ISave *isave) { 
	IOResult res;
	// Save common stuff
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();
	if (useCurve) {
		isave->BeginChunk(USE_CURVE_CHUNK);
		isave->EndChunk();
		}
	isave->BeginChunk(VERS2_CHUNK);
	isave->EndChunk();
	return IO_OK;
	}	
	  
class MixPostLoad : public PostLoadCallback {
	public:
		Mix *tm;
		MixPostLoad(Mix *b) {tm=b;}
		void proc(ILoad *iload) {  
		    tm->SetCrvA(.3f, TimeValue(0));
			tm->SetCrvB(.7f, TimeValue(0));
			delete this; 
			} 
	};

IOResult Mix::Load(ILoad *iload) { 
//	ULONG nb;
	IOResult res;
	int version = 0;
	iload->RegisterPostLoadCallback(new ParamBlockPLCB(oldVersions,1, &curVersion, this,0));
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			case USE_CURVE_CHUNK:
				useCurve = TRUE;
				break;
			case VERS2_CHUNK:
				version = 2;
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	if (version<2) 
		iload->RegisterPostLoadCallback(new MixPostLoad(this));
	return IO_OK;
	}


void MixDlg::DrawCurve(HDC hdc, Rect& rect) {
	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);
	SelectPen(hdc,(HPEN)GetStockObject(BLACK_PEN));
	SelectPen(hdc,(HBRUSH)GetStockObject(WHITE_BRUSH));
	Rectangle(hdc,rect.left, rect.top, rect.right, rect.bottom);
	MoveToEx(hdc,rect.left+1, rect.bottom-1,NULL); 
	float fx,fy;
	int ix,iy;
	float w = (float)rect.w()-2;
	float h = (float)rect.h()-2;
	for (ix =0; ix<w; ix++) {
		fx = (float(ix)+0.5f)/w;
		fy = theTex->mixCurve(fx);
		iy = int(h*fy+0.5f);
		LineTo(hdc, rect.left+1+ix, rect.bottom-1-iy);
		}
	}

void MixDlg::UpdateCurve() {
	HDC hdc = GetDC(hwCurve);
	Rect r;
	GetClientRect(hwCurve,&r);
	DrawCurve(hdc, r);
	ReleaseDC(hwCurve,hdc);
	}


BOOL MixDlg::CurveWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
	switch(msg) {
		case WM_COMMAND: 	break;
		case WM_MOUSEMOVE: 		break;
		case WM_LBUTTONUP: 	break;
		case WM_PAINT: 	
			{
			PAINTSTRUCT ps;
			Rect rect;
			HDC hdc = BeginPaint( hwnd, &ps );
			if (!IsRectEmpty(&ps.rcPaint)) {
				GetClientRect( hwnd, &rect );
				DrawCurve(hdc, rect);
				}
			EndPaint( hwnd, &ps );
			}													
			break;
		case WM_CREATE:
		case WM_DESTROY: 
			break;
    	}
	return DefWindowProc(hwnd,msg,wParam,lParam);
	}

static LRESULT CALLBACK CurveWndProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
	HWND hwpar = GetParent(hwnd);
	MixDlg *theDlg = (MixDlg *)GetWindowLong(hwpar, GWL_USERDATA);
	if (theDlg==NULL) return FALSE;
	int	res = theDlg->CurveWindowProc(hwnd,msg,wParam,lParam);
	return res;
	} 

