/**********************************************************************
 *<
	FILE: mixmat.cpp

	DESCRIPTION:  A blend of two materials

	CREATED BY: Rolf Berteig

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include "stdmtl.h"
#include "simpobj.h"

#define A_RENDER			A_PLUGIN1


extern HINSTANCE hInstance;
static LRESULT CALLBACK CurveWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

static Class_ID mixmatClassID(MIXMAT_CLASS_ID+2,999);


class ParticleShade : public ShadeContext
	{
	public:
		ShadeContext *s;
		Point3 NewUVW;
		ParticleShade(ShadeContext *ss) {
			s = ss;
			mode= ss->mode;
			doMaps= ss->doMaps;
			filterMaps = ss->filterMaps;;
			shadow = ss->shadow;
			backFace = ss->backFace;
			mtlNum = ss->mtlNum;
			ambientLight = ss->ambientLight;
			nLights = ss->nLights;
			out = ss->out;
			NewUVW = ss->UVW();

			}
	LightDesc *Light(int i ) { return s->Light(i);};
	int CurTime() {	return s->CurTime();};
	int FaceNumber() {return  s->FaceNumber();};
	Point3 Normal() { return s->Normal();};
	Point3 GNormal(){ return s->GNormal();};
	Point3 ReflectVector(void) {return s->ReflectVector();};
	Point3 RefractVector(float i) {return s->RefractVector(i);};
	Point3 CamPos() {return s->CamPos();};
	Point3 V() {return s->V();};
	Point3 P() {return s->P();};
	Point3 DP() {return s->DP();};
	Point3 PObj() {return s->PObj();};
	Point3 DPObj() {return s->DPObj();};
	Box3 ObjectBox() {return s->ObjectBox();};
	Point3 PObjRelBox() {return s->PObjRelBox();};
	Point3 DPObjRelBox() {return s->DPObjRelBox();};
	void ScreenUV(Point2 &p1,Point2 &p2) {s->ScreenUV(p1,p2);};
	IPoint2 ScreenCoord() {return s->ScreenCoord();};
	Point3 UVW() {return NewUVW;};
	Point3 DUVW() {return s->DUVW();};
	void DPdUVW(Point3 p1[]) {DPdUVW(p1);};
	void GetBGColor(Color &p1,Color &p2,int p3) {s->GetBGColor(p1,p2,p3);};
	Point3 PointTo(const Point3 &p1,RefFrame p2) {return s->PointTo(p1,p2);};
	Point3 PointFrom(const Point3 &p1,RefFrame p2) {return s->PointFrom(p1,p2);};
	Point3 VectorTo(const Point3 &p1,RefFrame p2) {return s->VectorTo(p1,p2);};
	Point3 VectorFrom(const Point3 &p1,RefFrame p2) {return s->VectorFrom(p1,p2);};

	int InMtlEditor() {return s->InMtlEditor();};
	void SetView(Point3 p1) {s->SetView(p1);};
//	Point3 UVW(int i ) { return s->UVW(i);};
	Point3 UVW(int i ) { return NewUVW;};
	Point3 DUVW(int i ) { return s->DUVW(i);};
	void DPdUVW(class Point3 p1[],int i) { s->DPdUVW(p1,i);};


	};

class MixMat;

class MixMatDlg : public ParamDlg {
	public:		
		HWND hwmedit;
		IMtlParams *ip;
		ICustButton *iPick;
		Interface *rip;
//		IRendParams *rip;

		MixMat *theMtl;
		HWND hPanel; 
		ISpinnerControl *iPersistance,*iSize, *crvASpin, *crvBSpin;
		ISpinnerControl *iEffectSize;
		BOOL valid;

		MixMatDlg(HWND hwMtlEdit, IMtlParams *imp, MixMat *m); 
		~MixMatDlg();
		
		void UpdateCurve();
		void DrawCurve(HDC hdc, Rect& rect);
		BOOL CurveWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
		BOOL WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);				
		void Invalidate();		
		void SetObjectName(HWND hWnd);
		// methods inherited from ParamDlg:		
		void ReloadDialog();
		Class_ID ClassID() {return mixmatClassID;}
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theMtl; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t) {Invalidate();}
		void ActivateDlg(BOOL onOff) {}


//		void Remove();
//		BOOL Pick(INode *node);
//		BOOL Filter(INode *node);
//		void EnterMode() {iPick->SetCheck(TRUE);}
//		void ExitMode() {iPick->SetCheck(FALSE);}

	};

#define PB_REF		0
#define SUB1_REF	1
#define SUB2_REF	2
//#define MAP_REF		3
#define PARTICLE_REF 3

#define PB_PERSISTANCE 0
#define PB_CURVEA	1
#define PB_CURVEB	2
#define PB_USECURVE	3
#define PB_USEMAT	4
#define PB_SIZE		5
#define PB_LOCALIZE 6
#define PB_XYZ		7
#define PB_HOSE		8
#define PB_PSIZE	9


class MixMat : public Mtl {	
	public:
		MixMatDlg *dlg;
		IParamBlock *pblock; 	// ref #0
		Mtl *sub1, *sub2;		// ref #1, 2		
//		Texmap *map;			// ref #3
		float u, crvA, crvB, esize;
		BOOL useCurve;		
		Interval ivalid;
		BOOL Localize;
		BOOL Hose;
		BOOL usePSize;
		int XYZ;

		Tab<Point3> PList;
		Tab<TimeValue> TList;
		Tab<float> SList;

		float persistance;//,Size;
		INode *node;	  // Ref #4
//		SimpleParticle *particles;

		IObjParam *Tempip;


		MixMat(BOOL loading);
		void NotifyChanged() {NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);}
		Mtl *UseMtl();
		float mixCurve(float x);

		// From MtlBase and Mtl
		void SetAmbient(Color c, TimeValue t) {}		
		void SetDiffuse(Color c, TimeValue t) {}		
		void SetSpecular(Color c, TimeValue t) {}
		void SetShininess(float v, TimeValue t) {}				
		
		Color GetAmbient(int mtlNum=0, BOOL backFace=FALSE);
	    Color GetDiffuse(int mtlNum=0, BOOL backFace=FALSE);
		Color GetSpecular(int mtlNum=0, BOOL backFace=FALSE);
		float GetXParency(int mtlNum=0, BOOL backFace=FALSE);
		float GetShininess(int mtlNum=0, BOOL backFace=FALSE);		
		float GetShinStr(int mtlNum=0, BOOL backFace=FALSE);
		float WireSize(int mtlNum=0, BOOL backFace=FALSE);
				
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		
		void Shade(ShadeContext& sc);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t);
		
		Class_ID ClassID() {return mixmatClassID; }
		SClass_ID SuperClassID() {return MATERIAL_CLASS_ID;}
		void GetClassName(TSTR& s) {s=GetString(IDS_RB_BLENDMAT);}  

		void DeleteThis() {delete this;}	

		// Methods to access sub-materials of meta-materials
	   	int NumSubMtls() {return 2;}
		Mtl* GetSubMtl(int i) {return i?sub2:sub1;}
		void SetSubMtl(int i, Mtl *m) {ReplaceReference(i+1,m);}
		TSTR GetSubMtlSlotName(int i) {return i?GetString(IDS_RB_BACK):GetString(IDS_RB_FACING);}

		// Methods to access sub texture maps of material or texmap
		int NumSubTexmaps() {return 1;}
//		Texmap* GetSubTexmap(int i) {return map;}
//		void SetSubTexmap(int i, Texmap *m) {ReplaceReference(MAP_REF,m);}
//		TSTR GetSubTexmapSlotName(int i) {return GetString(IDS_RB_MASK);}
//		TSTR GetSubTexmapTVName(int i) {return GetString(IDS_RB_MASK);}

		int NumSubs() {return 3;} 
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) {return subNum;}

		// From ref
 		int NumRefs() {return 4;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message);

		// IO
		IOResult Save(ISave *isave); 
		IOResult Load(ILoad *iload); 
	};


class MixMatClassDesc:public ClassDesc {
	public:
	int 			IsPublic() {return 1;}
	void *			Create(BOOL loading) {return new MixMat(loading);}
	const TCHAR *	ClassName() {return GetString(IDS_RB_PPAINT);}
	SClass_ID		SuperClassID() {return MATERIAL_CLASS_ID;}
	Class_ID 		ClassID() {return mixmatClassID;}
	const TCHAR* 	Category() {return _T("");}
	};
static MixMatClassDesc mixmatCD;

ClassDesc* GetMixMatDesc() {return &mixmatCD;}


class ParticleDialog : public HitByNameDlgCallback {
public:
	MixMat *Mix;
	ParticleDialog(MixMat *p) {Mix=p;};
	TCHAR *dialogTitle() {return _T("Select ParticleSystem");};
	TCHAR *buttonText() {return _T("Select");};
	BOOL singleSelect() {return TRUE;};
	BOOL useProc() {return TRUE;};
	void proc(INodeTab &nodeTab);
	BOOL useFilter() {return TRUE;};
	int filter(INode *node);
	};


int ParticleDialog::filter(INode *node)

{
	ObjectState os = node->EvalWorldState(0);
	if (os.obj->GetInterface(I_PARTICLEOBJ)==NULL)
		return FALSE;
		else return TRUE;

//	return (os.obj->ClassID()==MGRAPH_CLASSID);
//	return TRUE;
//	else return FALSE;

}

void ParticleDialog::proc(INodeTab &nodeTab)

{
  Mix->ReplaceReference(PARTICLE_REF,nodeTab[0]);
// po->ObjectNode = nodeTab[0];
}


//--- MixMatDlg ------------------------------------------------------
/*
BOOL MixMatDlg::Pick(INode *node)
	{
	theMtl->ReplaceReference(PARTICLE_REF,node);
//	SetObjectName(hWnd);
	return TRUE;
	}
*/

static BOOL CALLBACK PanelDlgProc(
		HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
	MixMatDlg *dlg = (MixMatDlg*)GetWindowLong(hWnd,GWL_USERDATA);
	switch (msg) {
		case WM_INITDIALOG:
			dlg = (MixMatDlg*)lParam;
			SetWindowLong(hWnd,GWL_USERDATA,lParam);
			break;
		}
	if (dlg) return dlg->WndProc(hWnd,msg,wParam,lParam);
	else return FALSE;
	}


MixMatDlg::MixMatDlg(HWND hwMtlEdit, IMtlParams *imp, MixMat *m)
	{
	hwmedit = hwMtlEdit;
	ip      = imp;
	theMtl  = m;
	valid   = FALSE;
	iPersistance    = NULL;
	iEffectSize    = NULL;
	iPick   = NULL;

	hPanel = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_MIXMAT),
		PanelDlgProc,
		GetString(IDS_DS_BASIC),
		(LPARAM)this);
	}

MixMatDlg::~MixMatDlg()
	{
	theMtl->dlg = NULL;	
	ReleaseISpinner(iPersistance);
	ReleaseISpinner(iEffectSize);
	ReleaseISpinner(crvASpin);
	ReleaseISpinner(crvBSpin);
	ReleaseICustButton(iPick);

	SetWindowLong(hPanel, GWL_USERDATA, NULL);
	}


void MixMatDlg::SetThing(ReferenceTarget *m) {
	theMtl = (MixMat*)m;
	ReloadDialog();
	}

void MixMatDlg::SetObjectName(HWND hWnd)
	{ 	
 	if (theMtl->node!=NULL){
		TSTR fname;
		fname = theMtl->node->GetName();
		SetDlgItemText(hWnd, IDC_PARTICLE,fname);
	} else {
		SetDlgItemText(hWnd, IDC_PARTICLE, GetString(IDS_RB_NONE));
		}
	}


BOOL MixMatDlg::WndProc(
		HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
	switch (msg) {
		case WM_INITDIALOG:

			iPersistance = GetISpinner(GetDlgItem(hWnd,IDC_PERSISTANCE_SPIN));
			iPersistance->SetLimits(0.0f,99999.0f,FALSE);
			iPersistance->SetScale(1.0f);
			iPersistance->LinkToEdit(GetDlgItem(hWnd,IDC_PERSISTANCE),EDITTYPE_FLOAT);

			iEffectSize = GetISpinner(GetDlgItem(hWnd,IDC_ESIZE_SPIN));
			iEffectSize->SetLimits(0.0f,99999.0f,FALSE);
			iEffectSize->SetScale(1.0f);
			iEffectSize->LinkToEdit(GetDlgItem(hWnd,IDC_ESIZE),EDITTYPE_FLOAT);

			crvASpin = SetupFloatSpinner(hWnd, IDC_MIXA_SPIN, IDC_MIXA_EDIT,0.0f,1.0f,.25f,.01f);
			crvBSpin = SetupFloatSpinner(hWnd, IDC_MIXB_SPIN, IDC_MIXB_EDIT,0.0f,1.0f,.75f,.01f);			
			SetWindowLong(GetDlgItem(hWnd, IDC_MIXCURVE), GWL_WNDPROC, (LONG)CurveWndProc);
			SetObjectName(hWnd);
			break;			

		case WM_PAINT:
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			return FALSE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_PARTICLE:
					Interface *rip;
					rip = GetCOREInterface();
					rip->DoHitByNameDialog(new ParticleDialog(theMtl));
					SetObjectName(hWnd);
					break;
				case IDC_MIXMAT_MAT1:
					PostMessage(hwmedit,WM_SUB_MTL_BUTTON, 
						0 ,(LPARAM)theMtl);
					break;
				case IDC_MIXMAT_MAT2:
					PostMessage(hwmedit,WM_SUB_MTL_BUTTON, 
						1 ,(LPARAM)theMtl);
					break;
				
				case IDC_MIXMAT_MAP:
					PostMessage(hwmedit,WM_TEXMAP_BUTTON, 
						0 ,(LPARAM)theMtl);	
					break;

				case IDC_MIX_USECURVE:
					theMtl->pblock->SetValue(
						PB_USECURVE,ip->GetTime(),
						(int)IsDlgButtonChecked(hWnd,IDC_MIX_USECURVE));
					ip->MtlChanged();
					break;

				case IDC_LOCALIZE:
					theMtl->pblock->SetValue(
						PB_LOCALIZE,ip->GetTime(),
						(int)IsDlgButtonChecked(hWnd,IDC_LOCALIZE));
					break;

				case IDC_HOSE:
					theMtl->pblock->SetValue(
						PB_HOSE,ip->GetTime(),
						(int)IsDlgButtonChecked(hWnd,IDC_HOSE));
					break;
				case IDC_USER_PARTICLE_SIZE:
					theMtl->pblock->SetValue(
						PB_PSIZE,ip->GetTime(),
						(int)IsDlgButtonChecked(hWnd,IDC_USER_PARTICLE_SIZE));
					break;


				case IDC_MIXMAT_USE1:
				case IDC_MIXMAT_USE2:
					theMtl->pblock->SetValue(
						PB_USEMAT,0,LOWORD(wParam)-IDC_MIXMAT_USE1);
					theMtl->NotifyChanged();
		    		ip->MtlChanged();
					break;

				case IDC_X:
				case IDC_Y:
				case IDC_Z:
					theMtl->pblock->SetValue(
						PB_XYZ,0,LOWORD(wParam)-IDC_X);
					break;


				}
			break;
		
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hWnd,msg,wParam,lParam);
			return FALSE;
		
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (LOWORD(wParam)) {
//				case IDC_MIXMAT_MIXSPIN:					
				case IDC_PERSISTANCE_SPIN:					
 					theMtl->pblock->SetValue(
 						PB_PERSISTANCE,ip->GetTime(),iPersistance->GetFVal());
 					break; 				
				case IDC_ESIZE_SPIN:					
 					theMtl->pblock->SetValue(
 						PB_SIZE,ip->GetTime(),iEffectSize->GetFVal());
 					break; 				
				case IDC_MIXA_SPIN: {					
					theMtl->pblock->SetValue(
						PB_CURVEA,ip->GetTime(),crvASpin->GetFVal());  
					float b;
					theMtl->pblock->GetValue(PB_CURVEB,ip->GetTime(),b,FOREVER);					
					if (b<crvASpin->GetFVal()) {
						theMtl->pblock->SetValue(
							PB_CURVEB,ip->GetTime(),crvASpin->GetFVal());
						}
					UpdateCurve();
					break;
					}
				case IDC_MIXB_SPIN: {
					theMtl->pblock->SetValue(
						PB_CURVEB,ip->GetTime(),crvBSpin->GetFVal());  
					float a;
					theMtl->pblock->GetValue(PB_CURVEA,ip->GetTime(),a,FOREVER);
					if (a>crvBSpin->GetFVal()) {
						theMtl->pblock->SetValue(
							PB_CURVEA,ip->GetTime(),crvBSpin->GetFVal());
						}
					UpdateCurve();
					break;
					}				
				}
			break;

		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;

		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			theMtl->NotifyChanged();
		    ip->MtlChanged();
			break;

		default:
			return FALSE;
		}
	return TRUE;
	}

void MixMatDlg::Invalidate()
	{
	valid = FALSE;
	InvalidateRect(hPanel,NULL,FALSE);	
	}
		
void MixMatDlg::ReloadDialog()
	{
	float u;
	int b;
	
	theMtl->pblock->GetValue(PB_PERSISTANCE,ip->GetTime(),u,FOREVER);
	iPersistance->SetValue(u,FALSE);

	theMtl->pblock->GetValue(PB_SIZE,ip->GetTime(),u,FOREVER);
	iEffectSize->SetValue(u,FALSE);

	theMtl->pblock->GetValue(PB_CURVEA,ip->GetTime(),u,FOREVER);
	crvASpin->SetValue(u,FALSE);
	theMtl->pblock->GetValue(PB_CURVEB,ip->GetTime(),u,FOREVER);
	crvBSpin->SetValue(u,FALSE);
	theMtl->pblock->GetValue(PB_USECURVE,ip->GetTime(),b,FOREVER);
	CheckDlgButton(hPanel,IDC_MIX_USECURVE,b);
	theMtl->pblock->GetValue(PB_LOCALIZE,ip->GetTime(),b,FOREVER);
	CheckDlgButton(hPanel,IDC_LOCALIZE,b);

	theMtl->pblock->GetValue(PB_HOSE,ip->GetTime(),b,FOREVER);
	CheckDlgButton(hPanel,IDC_HOSE,b);

	theMtl->pblock->GetValue(PB_PSIZE,ip->GetTime(),b,FOREVER);
	CheckDlgButton(hPanel,IDC_USER_PARTICLE_SIZE,b);

	theMtl->pblock->GetValue(PB_USEMAT,ip->GetTime(),b,FOREVER);
	CheckRadioButton(hPanel,IDC_MIXMAT_USE1,IDC_MIXMAT_USE2,IDC_MIXMAT_USE1+b);

	theMtl->pblock->GetValue(PB_XYZ,ip->GetTime(),b,FOREVER);
	CheckRadioButton(hPanel,IDC_X,IDC_Z,IDC_X+b);


	TSTR name;
	if (theMtl->sub1) name = theMtl->sub1->GetFullName();
	else name = GetString(IDS_DS_NONE);
	SetDlgItemText(hPanel, IDC_MIXMAT_MAT1, name);
	if (theMtl->sub2) name = theMtl->sub2->GetFullName();
	else name = GetString(IDS_RB_NONE);
	SetDlgItemText(hPanel, IDC_MIXMAT_MAT2, name);
//	if (theMtl->map) name = theMtl->map->GetFullName();
//	else name = GetString(IDS_RB_NONE);

/*
	SetDlgItemText(hPanel, IDC_MIXMAT_MAP, name);
	
	if (theMtl->map) {
		iPersistance->Disable();
	} else {
		iPersistance->Enable();
		}
*/
	UpdateCurve();

	}

void MixMatDlg::DrawCurve(HDC hdc, Rect& rect) {
	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);
	SelectPen(hdc,(HPEN)GetStockObject(BLACK_PEN));
	SelectPen(hdc,(HBRUSH)GetStockObject(WHITE_BRUSH));
	Rectangle(hdc,rect.left, rect.top, rect.right, rect.bottom);
	MoveToEx(hdc,rect.left+1, rect.bottom-1,NULL); 
	float fx,fy;
	int ix,iy;
	float w = (float)rect.w()-2;
	float h = (float)rect.h()-2;
	for (ix =0; ix<w; ix++) {
		fx = (float(ix)+0.5f)/w;
		fy = theMtl->mixCurve(fx);
		iy = int(h*fy+0.5f);
		LineTo(hdc, rect.left+1+ix, rect.bottom-1-iy);
		}
	}

void MixMatDlg::UpdateCurve() {
	theMtl->Update(ip->GetTime(),FOREVER);
	HWND hwCurve = GetDlgItem(hPanel,IDC_MIXCURVE); 
	HDC hdc = GetDC(hwCurve);
	Rect r;
	GetClientRect(hwCurve,&r);
	DrawCurve(hdc, r);
	ReleaseDC(hwCurve,hdc);
	}


BOOL MixMatDlg::CurveWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
	switch(msg) {		
		case WM_PAINT: {
			PAINTSTRUCT ps;
			Rect rect;
			HDC hdc = BeginPaint( hwnd, &ps );
			if (!IsRectEmpty(&ps.rcPaint)) {
				GetClientRect(hwnd, &rect);
				DrawCurve(hdc, rect);
				}
			EndPaint(hwnd, &ps);
			break;
			}													
    	}
	return DefWindowProc(hwnd,msg,wParam,lParam);
	}

static LRESULT CALLBACK CurveWndProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
	HWND hwpar = GetParent(hwnd);
	MixMatDlg *theDlg = (MixMatDlg *)GetWindowLong(hwpar, GWL_USERDATA);
	if (theDlg==NULL) return FALSE;	
	int	res = theDlg->CurveWindowProc(hwnd,msg,wParam,lParam);	
	return res;
	}


//--- DoubleSeded Material -------------------------------------------------

static ParamBlockDescID pbdesc0[] = {
	{TYPE_FLOAT, NULL, TRUE,0 },	// Mix Amount
	{TYPE_FLOAT, NULL, TRUE,1 },	// Curve A	
	{TYPE_FLOAT, NULL, TRUE,2 },  // Curve B
	{TYPE_INT, NULL, FALSE,3 },	// Use curve
	{TYPE_INT, NULL, FALSE,4 },   // Use mat
	{TYPE_FLOAT, NULL, TRUE,5 },	// Size
	{TYPE_INT, NULL, FALSE,6 },	// Localize
	{TYPE_INT, NULL, FALSE,7 },	// XYZ
	{TYPE_INT, NULL, FALSE,8 },	// Hose


};  

static ParamBlockDescID pbdesc1[] = {
	{TYPE_FLOAT, NULL, TRUE,0 },	// Mix Amount
	{TYPE_FLOAT, NULL, TRUE,1 },	// Curve A	
	{TYPE_FLOAT, NULL, TRUE,2 },  // Curve B
	{TYPE_INT, NULL, FALSE,3 },	// Use curve
	{TYPE_INT, NULL, FALSE,4 },   // Use mat
	{TYPE_FLOAT, NULL, TRUE,5 },	// Size
	{TYPE_INT, NULL, FALSE,6 },	// Localize
	{TYPE_INT, NULL, FALSE,7 },	// XYZ
	{TYPE_INT, NULL, FALSE,8 },	// Hose
	{TYPE_INT, NULL, FALSE,9 },	// Hose


};  

#define PBLOCK_LENGTH	10
// Array of old versions
static ParamVersionDesc versions[] = {
	ParamVersionDesc(pbdesc0,9,0)
	};
#define NUM_OLDVERSIONS	1

// Current version
#define CURRENT_VERSION	1
static ParamVersionDesc curVersion(pbdesc1,PBLOCK_LENGTH,CURRENT_VERSION);


MixMat::MixMat(BOOL loading)
	{	
	pblock = NULL;
	dlg = NULL;
	sub1 = sub2 = NULL;	
//	map = NULL;
	node = NULL;
	ivalid.SetEmpty();
	if (!loading) {
		Reset();
		}
	}

void MixMat::Reset()
	{
	MakeRefByID(FOREVER, PB_REF, 
		CreateParameterBlock(pbdesc1, PBLOCK_LENGTH, CURRENT_VERSION));
//	ReplaceReference(PB_REF,CreateParameterBlock(pbdesc1,PBLOCK_LENGTH));
	pblock->SetValue(PB_CURVEA,0,0.25f);
	pblock->SetValue(PB_CURVEB,0,0.75f);	
	pblock->SetValue(PB_SIZE,0,30.0f);	
	pblock->SetValue(PB_PERSISTANCE,0,10.0f);	

//	ReplaceReference(SUB1_REF,new StdMtl(FALSE));
//	ReplaceReference(SUB2_REF,new StdMtl(FALSE));
//	DeleteReference(MAP_REF);
	DeleteReference(PARTICLE_REF);
	}

Mtl *MixMat::UseMtl() 
	{
	int m;
	pblock->GetValue(PB_USEMAT,0,m,FOREVER);
	if (m==0 && sub1) return sub1;
	if (m==1 && sub2) return sub2;
	return sub1?sub1:sub2;
	}

Color MixMat::GetAmbient(int mtlNum, BOOL backFace) { 
	return UseMtl()?UseMtl()->GetAmbient(mtlNum,backFace):Color(0,0,0);
	}		
Color MixMat::GetDiffuse(int mtlNum, BOOL backFace){ 
	return UseMtl()?UseMtl()->GetDiffuse(mtlNum,backFace):Color(0,0,0);
	}				
Color MixMat::GetSpecular(int mtlNum, BOOL backFace){
	return UseMtl()?UseMtl()->GetSpecular(mtlNum,backFace):Color(0,0,0);
	}		
float MixMat::GetXParency(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}
float MixMat::GetShininess(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}		
float MixMat::GetShinStr(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->GetXParency(mtlNum,backFace):0.0f;
	}

float MixMat::WireSize(int mtlNum, BOOL backFace) {
	return UseMtl()?UseMtl()->WireSize(mtlNum,backFace):0.0f;
	}
		
ParamDlg* MixMat::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp)
	{
	dlg = new MixMatDlg(hwMtlEdit, imp, this);
	return dlg;
	}

float MixMat::mixCurve(float x) 
	{
	if (x<crvA)  return 0.0f;
	if (x>=crvB) return 1.0f;
	x = (x-crvA)/(crvB-crvA);
	return (x*x*(3-2*x));
	}

static Color black(0,0,0);

void MixMat::Shade(ShadeContext& sc) {

	ParticleShade PShade(&sc);

	if (gbufID) sc.SetGBufferID(gbufID);
	float mix = u;
//	if (map && useCurve) mix = mixCurve(mix);

	mix = 0.0f;
//loop through particles finding any within range

	INode *RNode = sc.Node();
	TSTR fname,MNode("MNode");
	fname = RNode->GetName();

	PShade.ResetOutput();
	if ((Localize) && (Hose))
		{
		if ( (node != NULL) && (!(fname == MNode)) )


			{
			Point3 ShadePoint,ShadePointWorld;
			ShadePoint = sc.P();
			ShadePointWorld = sc.PointTo(ShadePoint, REF_WORLD);
			mix = 0.0f;
			if (sub1) sub1->Shade(sc);

			for (int i=0; i<PList.Count(); i++) {
//check distance
				float d = LengthSquared(PList[i]-ShadePointWorld);
				float psize = esize;
				if (usePSize) psize = SList[i] * SList[i];
				if  (d < psize)
					{
					mix = 0.0f;
					Point3 TUVW;
					float per = d/psize;
					TUVW = Normalize(ShadePointWorld - PList[i]) * per;
//normalize;
//get projection direction
					if (XYZ == 0)
						{
						TUVW.x = (TUVW.y - 1.0f) / 2.0f;
						TUVW.y = (TUVW.z - 1.0f) / 2.0f;
						TUVW.z = 0.0f;
						}
					else if (XYZ == 1)
						{
						TUVW.x = (TUVW.x + 1.0f) / 2.0f;
						TUVW.y = (TUVW.z - 1.0f) / 2.0f;
						TUVW.z = 0.0f;
						}
					else if (XYZ == 2)
						{
						TUVW.x = (TUVW.x + 1.0f) / 2.0f;
						TUVW.y = (TUVW.y - 1.0f) / 2.0f;
						TUVW.z = 0.0f;
	
						}
			
					PShade.NewUVW = TUVW;
					if(sub2) 
						{
						sub2->Shade(PShade);
						mix = (PShade.out.t.r + PShade.out.t.g + PShade.out.t.b)/3.0f;
						PShade.out.t.r = 0.0f;
						PShade.out.t.g = 0.0f;
						PShade.out.t.b = 0.0f;
						sc.out.MixIn(PShade.out,mix);

		

						}



					}
//watch for age change	
				}



			}
		

		}
	else
		{
	
		if ( (node != NULL) && (!(fname == MNode)) )


			{
			Point3 ShadePoint,ShadePointWorld;
			ShadePoint = sc.P();
			ShadePointWorld = sc.PointTo(ShadePoint, REF_WORLD);
			mix = 0.0f;
			int closest = -1;
			float closestdist = -1.0f;
			TimeValue currentage = -99999;
			float psize = esize;
			for (int i=0; i<PList.Count(); i++) {
//check distance
				float d = LengthSquared(PList[i]-ShadePointWorld);
				psize = esize;
				if (usePSize) psize = SList[i] * SList[i];
				if ( (d < psize) && (!Localize))
					{
					float tmix;
					tmix = (psize-d)/psize;
					if (useCurve)
						mix += mixCurve(tmix);
					else mix += tmix;
					}
				else if (Localize)
					{
					if (d < psize)
						{

						if ((d<closestdist) || (closest == -1))
							{
							closest = i;
							closestdist = d;
							}
						}

					}
//watch for age change	
				}


			if ((Localize) && (closest != -1))
				{
				Point3 TUVW;
				float per = closestdist/psize;
				TUVW = Normalize(ShadePointWorld - PList[closest]) * per;
//normalize;
//get projection direction
				if (XYZ == 0)
					{
					TUVW.x = (TUVW.y - 1.0f) / 2.0f;
					TUVW.y = (TUVW.z - 1.0f) / 2.0f;
					TUVW.z = 0.0f;
					}
				else if (XYZ == 1)
					{
					TUVW.x = (TUVW.x + 1.0f) / 2.0f;
					TUVW.y = (TUVW.z - 1.0f) / 2.0f;
					TUVW.z = 0.0f;
					}
				else if (XYZ == 2)
					{
					TUVW.x = (TUVW.x + 1.0f) / 2.0f;
					TUVW.y = (TUVW.y - 1.0f) / 2.0f;
					TUVW.z = 0.0f;

					}
			
				PShade.NewUVW = TUVW;
				float tmix;
			
				tmix = (closestdist)/psize;

				if (useCurve)
					mix = mixCurve(tmix);
				else mix = tmix;


				}


			}
		

		if (mix==0.0f) {
			if (sub1) sub1->Shade(sc);
		} else
		if (mix>=1.0f) {

			if ((sub2) && (!Localize))
				{
				sub2->Shade(sc);
				}
			else if (Localize)
				{
				sub2->Shade(PShade);
				sc.out = PShade.out;
				}
		} else {
			if (sub1) {
				sub1->Shade(sc);
				if(sub2) {

					if (!Localize)
						{
						ShadeOutput out1 = sc.out;
						sc.ResetOutput();
						sub2->Shade(sc);
						sc.out.MixIn(out1,mix);
						}
					else
						{
						sub2->Shade(PShade);
						sc.out.MixIn(PShade.out,mix);

						}
					}
				}
			else {
				if (sub2) 
					{
					sub2->Shade(sc);
					}
				}
			}
		}
	}

void MixMat::Update(TimeValue t, Interval& valid)
	{	
	ivalid = FOREVER;
	if (sub1) sub1->Update(t,valid);
	if (sub2) sub2->Update(t,valid);
//	if (map) map->Update(t,valid);
//	if (node) node->Update(t,valid);

	pblock->GetValue(PB_PERSISTANCE,t,u,ivalid);
	pblock->GetValue(PB_PERSISTANCE,t,persistance,ivalid);
	pblock->GetValue(PB_SIZE,t,esize,ivalid);
	esize = esize * esize;
	pblock->GetValue(PB_CURVEA,t,crvA,ivalid);
	pblock->GetValue(PB_CURVEB,t,crvB,ivalid);
	pblock->GetValue(PB_USECURVE,t,useCurve,ivalid);
	pblock->GetValue(PB_LOCALIZE,t,Localize,ivalid);
	pblock->GetValue(PB_HOSE,t,Hose,ivalid);
	pblock->GetValue(PB_PSIZE,t,usePSize,ivalid);
	pblock->GetValue(PB_XYZ,t,XYZ,ivalid);
	valid &= ivalid;

//loop through particles finding any within range

	PList.ZeroCount();
	TList.ZeroCount();
	SList.ZeroCount();
	if (node != NULL)
		{
		SimpleParticle *particles;
		TimeValue CTime = (int) ((float)t-(persistance*160.0f));

		while (CTime < t)
			{
			ObjectState os = node->EvalWorldState(CTime);
			particles = (SimpleParticle*) os.obj;
			particles->Update(CTime,node);
			for (int i=0; i<particles->parts.Count(); i++) {
				TimeValue pt = particles->parts.ages[i];
				if (particles->parts.Alive(i))
					{
//check distance
					PList.Append(1,&particles->parts.points[i],1);
					TList.Append(1,&CTime,1);
					float s = particles->ParticleSize(CTime,i);
					SList.Append(1,&s,1);
					}
				}
			CTime += GetTicksPerFrame();
			}

		}

	}

Interval MixMat::Validity(TimeValue t)
	{
	Interval valid = FOREVER;		
	if (sub1) valid &= sub1->Validity(t);
	if (sub2) valid &= sub2->Validity(t);
//	if (map) valid &= map->Validity(t);
//	if (node) valid.SetInstant(t);

	pblock->GetValue(PB_PERSISTANCE,t,u,valid);
	pblock->GetValue(PB_SIZE,t,u,valid);
	pblock->GetValue(PB_CURVEA,t,u,valid);
	pblock->GetValue(PB_CURVEB,t,u,valid);
	return valid;
	}

Animatable* MixMat::SubAnim(int i)
	{
	switch (i) {
		case 0: return pblock;
		case 1: return sub1;
		case 2: return sub2;
//		case 3: return map;
		default: return NULL;
		}
	}

TSTR MixMat::SubAnimName(int i)
	{
	switch (i) {
		case 0: return GetString(IDS_DS_PARAMETERS);
		case 1: return GetString(IDS_RB_MATERIALONE);
		case 2: return GetString(IDS_RB_MATERIALTWO);
//		case 3: return GetString(IDS_RB_MASK);
		default: return _T("");
		}
	}

RefTargetHandle MixMat::GetReference(int i)
	{
	switch (i) {
		case 0: return pblock;
		case 1: return sub1;
		case 2: return sub2;
//		case 3: return map;
		case 3: return node;
		default: return NULL;
		}
	}

void MixMat::SetReference(int i, RefTargetHandle rtarg)
	{
	switch (i) {
		case 0: pblock = (IParamBlock*)rtarg; break;
		case 1: sub1 = (Mtl*)rtarg; break;
		case 2: sub2 = (Mtl*)rtarg; break;
//		case 3: map = (Texmap*)rtarg; break;
 		case 3: node = (INode*)rtarg; break;

		}
	}

RefTargetHandle MixMat::Clone(RemapDir &remap)
	{
	MixMat *mtl = new MixMat(FALSE);
	*((MtlBase*)mtl) = *((MtlBase*)this);  // copy superclass stuff
	mtl->ReplaceReference(PB_REF,remap.CloneRef(pblock));
	if (sub1) mtl->ReplaceReference(SUB1_REF,remap.CloneRef(sub1));
	if (sub2) mtl->ReplaceReference(SUB2_REF,remap.CloneRef(sub2));
//	if (map) mtl->ReplaceReference(MAP_REF,remap.CloneRef(map));
	if (node) mtl->ReplaceReference(PARTICLE_REF,node);
	return mtl;
	}

RefResult MixMat::NotifyRefChanged(
		Interval changeInt, 
		RefTargetHandle hTarget, 
		PartID& partID, 
		RefMessage message)
	{
	switch (message) {
		case REFMSG_CHANGE:
			if (hTarget==pblock) ivalid.SetEmpty();
			if (hTarget==node) 
				{
				ivalid.SetEmpty();
				return REF_SUCCEED;
				}
			if (dlg && dlg->theMtl==this) dlg->Invalidate();
			break;
		
		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_PERSISTANCE: gpd->dim = stdPercentDim; break;				
				//case PB_SIZE: gpd->dim = stdDim; break;				
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			switch (gpn->index) {
				case PB_PERSISTANCE: gpn->name = GetString(IDS_RB_MIXAMOUNT); break;
				case PB_SIZE: gpn->name = GetString(IDS_RB_ESIZE); break;
				case PB_CURVEA: gpn->name = GetString(IDS_RB_LOWER); break;
				case PB_CURVEB: gpn->name = GetString(IDS_RB_UPPER); break;
				}
			return REF_STOP; 
			}
		case REFMSG_TARGET_DELETED:
			if (hTarget==node) {
				node = NULL;
				}
			break;

		}
	return REF_SUCCEED;
	}


//
// Note: ALL Materials and texmaps must have a Save and Load to save and load
// the MtlBase info.
#define MTL_HDR_CHUNK 0x4000
IOResult MixMat::Save(ISave *isave) { 
	IOResult res;
	// Save common stuff
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();
	return IO_OK;
	}	
	  

IOResult MixMat::Load(ILoad *iload) { 
//	ULONG nb;
	IOResult res;

	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	iload->RegisterPostLoadCallback(
		new ParamBlockPLCB(versions,NUM_OLDVERSIONS,&curVersion,this,PB_REF));

	return IO_OK;
	}


