/**********************************************************************
 *<
	FILE:			Wood.cpp

	DESCRIPTION:	Wood 3D Texture map Class Decl.

	CREATED BY:		Suryan Stalin

	HISTORY:		Modified from Marble.cpp by adding IPAS wood stuff

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#ifndef __WOOD__H
#define __WOOD__H


//	defines
#define WOOD_CLASS_ID 		0x0000214
#define NSUBTEX				2
#define NCOLS				2
#define WOOD_VERSION		1
#define PB_SIZE				0
#define PB_R1				1
#define PB_R2				2
#define PB_COL1				3
#define PB_COL2				4
#define WOOD_VERS			0x3AE2
#define MTL_HDR_CHUNK		0x4000
#define WOODVERS1_CHUNK		0x4001
#define WD					.02f
#define FACT				500.0f
#define SZ					1.0f //50.0


#define LIM255(c)			(((c)>=255)?255:(c))
#define SIZE				1
#define R1					2
#define R2					3
#define COL1				10		
#define COL2				11
#define NOISE_DIM			20    
#define FNOISE_DIM			20.0


//	C Prototypes
ClassDesc*		GetWoodDesc();
static BOOL		CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam);


//	Class Declerations
class Wood;

class WoodDlg: public ParamDlg 
{
	public:
		HWND hwmedit;	 	// window handle of the materials editor dialog
		IMtlParams *ip;
		Wood *theTex;	// current Wood being edited.
		HWND hPanel; 		// Rollup pane
		ISpinnerControl *r1Spin,*r2Spin,*sizeSpin;
		IColorSwatch *cs[2];
		TimeValue curTime; 
		ParamDlg *xyzGenDlg;
		int isActive;
		//-----------------------------
		WoodDlg(HWND hwMtlEdit, IMtlParams *imp, Wood *m); 
		~WoodDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void UpdateSubTexNames();
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff);

		// methods inherited from ParamDlg:
		Class_ID ClassID();
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};

class Wood: public Tex3D 
{ 
	friend class WoodPostLoad;
	friend class WoodDlg;
	Color col[NCOLS];
	float r1,r2,size;
	XYZGen *xyzGen;		   // ref #0
	IParamBlock *pblock;   // ref #1
	Texmap* subTex[NSUBTEX];  // More refs
	Interval ivalid;
	int rollScroll;
	int vers;
	WoodDlg *paramDlg;
	float WoodFunc(float x, float y, float z);

	public:
		Wood();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }

		void SetColor(int i, Color c, TimeValue t);
		void SetSize(float f, TimeValue t);
		void SetR1(float f, TimeValue t);
		void SetR2(float f, TimeValue t);
		void NotifyChanged();
		void SwapInputs(); 

		void ReadSXPData(TCHAR *name, void *sxpdata);

		// Wood noise functions
		void	InitNoise();
		float	WoodNoise(float x);
		void	LerpColor(RGBA *c, RGBA *a, RGBA *b, float f);
		float	SmoothStep(float x0, float x1, float v);

		// Evaluate the color of map for the context.
		RGBA EvalColor(ShadeContext& sc);

		// For Bump mapping, need a perturbation to apply to a normal.
		// Leave it up to the Texmap to determine how to do this.
		Point3 EvalNormalPerturb(ShadeContext& sc);

		// Requirements
		ULONG LocalRequirements(int subMtlNum) { return xyzGen->Requirements(subMtlNum); }

		// Methods to access texture maps of material
		int NumSubTexmaps() { return NSUBTEX; }
		Texmap* GetSubTexmap(int i) { return subTex[i]; }
		void SetSubTexmap(int i, Texmap *m);
		TSTR GetSubTexmapSlotName(int i);

		Class_ID ClassID();
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_WOOD); }  
		void DeleteThis() { delete this; }	

		int NumSubs() { return 2+NSUBTEX; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return 2+NSUBTEX; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

};

class WoodClassDesc:public ClassDesc 
{
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new Wood; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_WOOD); }
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID();
	const TCHAR* 	Category() { return TEXMAP_CAT_3D;  }
};

class WoodPostLoad : public PostLoadCallback 
{
	public:
		Wood *chk;
		WoodPostLoad(Wood *b) {chk=b;}
		void proc(ILoad *iload) {
			if (chk->vers<1) {
				if (chk->pblock) 
					ScaleFloatController(chk->pblock, PB_SIZE, 100.0f);
//				iload->SetObsolete();
				}
			delete this;
			}
};


#pragma pack(1)
struct WoodState 
{
	ulong version;
	float size;
	float dx, dy, dz;
	float r1;
	float r2;
	Col24 col1,col2;
};
#pragma pack()

#endif
